#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/5/20 21:09
@File    : factor.py
@contact : mmmaaaggg@163.com
@desc    : 用来对时间序列数据建立因子
"""
import datetime
import logging
import bisect
import ffn
import numpy as np
import pandas as pd
import talib
from ibats_utils.mess import date_2_str

logger = logging.getLogger(__name__)
FFN_VERSION = ffn.__version__


def add_factor_of_trade_date(df: pd.DataFrame, trade_date_series):
    """
    自然日，交易日相关因子
    :param df:
    :param trade_date_series:
    :return:
    """
    # 自然日相关因子
    index_s = pd.Series(df.index)
    df['dayofweek'] = index_s.apply(lambda x: x.dayofweek).to_numpy()
    df['day'] = index_s.apply(lambda x: x.day).to_numpy()
    df['month'] = index_s.apply(lambda x: x.month).to_numpy()
    df['daysleftofmonth'] = index_s.apply(lambda x: x.days_in_month - x.day).to_numpy()

    # 本月第几个交易日, 本月还剩几个交易日
    groups = trade_date_series.groupby(trade_date_series.apply(lambda x: datetime.datetime(x.year, x.month, 1)))

    def get_td_of_month(dt):
        first_day_of_month = datetime.datetime(dt.year, dt.month, 1)
        month_s = groups.get_group(first_day_of_month)
        td_of_month = (month_s <= dt).sum()
        td_left_of_month = (month_s > dt).sum()
        return month_s.shape[0], td_of_month, td_left_of_month

    result_arr = index_s.apply(get_td_of_month).to_numpy()
    df['td_of_month'] = [_[0] for _ in result_arr]
    df['td_pass_of_month'] = [_[1] for _ in result_arr]
    df['td_left_of_month'] = [_[2] for _ in result_arr]

    # 本周第几个交易日, 本周还剩几个交易日
    groups = trade_date_series.groupby(trade_date_series.apply(lambda x: x.year * 100 + x.weekofyear))

    def get_td_of_week(dt):
        name = dt.year * 100 + dt.weekofyear
        week_s = groups.get_group(name)
        td_pass_of_week = (week_s <= dt).sum()
        td_left_of_week = (week_s > dt).sum()
        return week_s.shape[0], td_pass_of_week, td_left_of_week

    result_arr = index_s.apply(get_td_of_week).to_numpy()
    df['td_of_week'] = [_[0] for _ in result_arr]
    df['td_pass_of_week'] = [_[1] for _ in result_arr]
    df['td_left_of_week'] = [_[2] for _ in result_arr]

    # 距离下一次放假交易日数（超过2天以上的休息日）
    # 计算距离下一个交易日的日期
    days_2_next_trade_date_s = (trade_date_series.shift(-1) - trade_date_series).fillna(pd.Timedelta(days=0))
    days_2_next_trade_date_s.index = pd.DatetimeIndex(trade_date_series)
    # 倒序循环，计算距离下一次放假的日期
    days_count, trade_date_count, result_dic, first_date = np.nan, np.nan, {}, index_s[0]
    for trade_date, delta in days_2_next_trade_date_s.sort_index(ascending=False).items():
        if trade_date < first_date:
            break
        days = delta.days
        if days > 3:
            trade_date_count, days_count = 0, 0
        elif pd.isna(days_count) >= 0 and days >= 1:
            trade_date_count += 1
            days_count += days
        else:
            trade_date_count, days_count = np.nan, np.nan

        result_dic[trade_date] = [days_count, trade_date_count]

    vacation_df = pd.DataFrame(result_dic).T.sort_index().rename(columns={0: 'days_2_vacation', 1: 'td_2_vacation'})
    df = df.join(vacation_df, how='left')

    return df


def add_factor_of_delivery_date(df, delivery_date_series):
    index_s = pd.Series(df.index)

    # 当期合约距离交割日天数
    result_arr = []
    for _, trade_date in index_s.items():
        next_2_date = delivery_date_series[delivery_date_series > trade_date].head(2)
        if next_2_date.shape[0] < 2:
            day_2_first_del_date, day_2_second_del_date = np.nan, np.nan
        else:
            first_del_date, secend_del_date = next_2_date[0], next_2_date[1]
            day_2_first_del_date = (first_del_date - trade_date).days
            day_2_second_del_date = (secend_del_date - trade_date).days

        result_arr.append([day_2_first_del_date, day_2_second_del_date])

    df['days_2_first_del_date'] = [_[0] for _ in result_arr]
    df['days_2_second_del_date'] = [_[1] for _ in result_arr]
    return df


def add_factor_of_price(df: pd.DataFrame, ohlcav_col_name_list, drop=False, log_av=True):
    open_key = ohlcav_col_name_list[0]
    high_key = ohlcav_col_name_list[1]
    low_key = ohlcav_col_name_list[2]
    close_key = ohlcav_col_name_list[3]
    amount_key = ohlcav_col_name_list[4]
    volume_key = ohlcav_col_name_list[5]
    open_s = df[open_key]
    high_s = df[high_key]
    low_s = df[low_key]
    close_s = df[close_key]
    amount_s = df[amount_key]
    volume_s = df[volume_key]
    # 平均成交价格
    deal_price_s = amount_s / volume_s
    deal_price_s[volume_s.isna()] = ((open_s * 2 + high_s + low_s + close_s * 2) / 6)[volume_s.isna()]
    df[f'deal_price'] = deal_price_s
    # 均线因子
    df[f'rr'] = close_s.to_returns()
    for n in [5, 10, 15, 20, 30, 60]:
        df[f'ma{n}'] = close_s.rolling(n).mean()
    # 波动率因子
    expanding = close_s.expanding(5)
    df[f'volatility_all'] = expanding.std() / expanding.mean()
    for n in [20, 60]:
        df[f'volatility{n}'] = close_s.rolling(n).std() / close_s.rolling(n).mean()

    # 收益率方差
    rr = close_s.to_returns()
    for n in [20, 60]:
        df[f'rr_std{n}'] = rr.rolling(n).std()

    #累积/派发线（Accumulation / Distribution Line，该指标将每日的成交量通过价格加权累计，
    #用以计算成交量的动量。属于趋势型因子
    df['AD'] = talib.AD(high_s, low_s, close_s, volume_s)

    # 佳庆指标（Chaikin Oscillator），该指标基于AD曲线的指数移动均线而计算得到。属于趋势型因子
    df['ADOSC'] = talib.ADOSC(high_s, low_s, close_s, volume_s, fastperiod=3, slowperiod=10)

    # 平均动向指数，DMI因子的构成部分。属于趋势型因子
    df['ADX'] = talib.ADX(high_s, low_s, close_s, timeperiod=14)

    # 相对平均动向指数，DMI因子的构成部分。属于趋势型因子
    df['ADXR'] = talib.ADXR(high_s, low_s, close_s, timeperiod=14)

    # 绝对价格振荡指数
    df['APO'] = talib.APO(close_s, fastperiod=12, slowperiod=26)

    # Aroon通过计算自价格达到近期最高值和最低值以来所经过的期间数，
    # 帮助投资者预测证券价格从趋势到区域区域或反转的变化，
    # Aroon指标分为Aroon、AroonUp和AroonDown3个具体指标。属于趋势型因子
    df['AROONDown'], df['AROONUp'] = talib.AROON(high_s, low_s, timeperiod=14)
    df['AROONOSC'] = talib.AROONOSC(high_s, low_s, timeperiod=14)

    # 均幅指标（Average TRUE Ranger），取一定时间周期内的股价波动幅度的移动平均值，
    # 是显示市场变化率的指标，主要用于研判买卖时机。属于超买超卖型因子。
    for n in [6, 14]:
        df[f'ATR{n}'] = talib.ATR(high_s, low_s, close_s, timeperiod=n)

    # 布林带
    df['Boll_Up'], df['Boll_Mid'], df['Boll_Down'] = \
        talib.BBANDS(close_s, timeperiod=20, nbdevup=2, nbdevdn=2, matype=0)

    # 均势指标
    df['BOP'] = talib.BOP(open_s, high_s, low_s, close_s)

    # 5日顺势指标（Commodity Channel Index），专门测量股价是否已超出常态分布范围。属于超买超卖型因子。
    for n in [5, 10, 20, 88]:
        df[f'CCI{n}'] = talib.CCI(high_s, low_s, close_s, timeperiod=5)

    # 钱德动量摆动指标（Chande Momentum Osciliator），与其他动量指标摆动指标如
    # 相对强弱指标（RSI）和随机指标（KDJ）不同，
    # 钱德动量指标在计算公式的分子中采用上涨日和下跌日的数据。属于超买超卖型因子
    df['CMO_Close'] = talib.CMO(close_s, timeperiod=14)
    df['CMO_Open'] = talib.CMO(open_s, timeperiod=14)

    # DEMA双指数移动平均线
    for n in [6, 12, 26]:
        df[f'DEMA{n}'] = talib.DEMA(close_s, timeperiod=n)

    # DX 动向指数
    df['DX'] = talib.DX(high_s, low_s, close_s, timeperiod=14)

    # EMA 指数移动平均线
    for n in [6, 12, 26, 60]:
        df[f'EMA{n}'] = talib.EMA(close_s, timeperiod=n)

    # KAMA 适应性移动平均线
    df['KAMA'] = talib.KAMA(close_s, timeperiod=30)

    # MACD
    df['MACD_DIF'], df['MACD_DEA'], df['MACD_bar'] = \
        talib.MACD(close_s, fastperiod=12, slowperiod=24, signalperiod=9)

    # 中位数价格 不知道是什么意思
    df['MEDPRICE'] = talib.MEDPRICE(high_s, low_s)

    # 负向指标 负向运动
    df['MiNUS_DI'] = talib.MINUS_DI(high_s, low_s, close_s, timeperiod=14)
    df['MiNUS_DM'] = talib.MINUS_DM(high_s, low_s, timeperiod=14)

    # 动量指标（Momentom Index），动量指数以分析股价波动的速度为目的，研究股价在波动过程中各种加速，
    # 减速，惯性作用以及股价由静到动或由动转静的现象。属于趋势型因子
    df['MOM'] = talib.MOM(close_s, timeperiod=10)

    # 归一化平均值范围
    df['NATR'] = talib.NATR(high_s, low_s, close_s, timeperiod=14)

    # OBV 	能量潮指标（On Balance Volume，OBV），以股市的成交量变化来衡量股市的推动力，
    # 从而研判股价的走势。属于成交量型因子
    df['OBV'] = talib.OBV(close_s, volume_s)

    # PLUS_DI 更向指示器
    df['PLUS_DI'] = talib.PLUS_DI(high_s, low_s, close_s, timeperiod=14)
    df['PLUS_DM'] = talib.PLUS_DM(high_s, low_s, timeperiod=14)

    # PPO 价格振荡百分比
    df['PPO'] = talib.PPO(close_s, fastperiod=6, slowperiod=26, matype=0)

    # ROC 6日变动速率（Price Rate of Change），以当日的收盘价和N天前的收盘价比较，
    # 通过计算股价某一段时间内收盘价变动的比例，应用价格的移动比较来测量价位动量。属于超买超卖型因子。
    for n in [6, 20]:
        df[f'ROC{n}'] = talib.ROC(close_s, timeperiod=n)
    # 12日量变动速率指标（Volume Rate of Change），以今天的成交量和N天前的成交量比较，
    # 通过计算某一段时间内成交量变动的幅度，应用成交量的移动比较来测量成交量运动趋向，
    # 达到事先探测成交量供需的强弱，进而分析成交量的发展趋势及其将来是否有转势的意愿，
    # 属于成交量的反趋向指标。属于成交量型因子
    for n in [6, 20]:
        df[f'VROC{n}'] = talib.ROC(volume_s, timeperiod=n)

    # ROC 6日变动速率（Price Rate of Change），以当日的收盘价和N天前的收盘价比较，
    # 通过计算股价某一段时间内收盘价变动的比例，应用价格的移动比较来测量价位动量。属于超买超卖型因子。
    for n in [6, 20]:
        df[f'ROCP{n}'] = talib.ROCP(close_s, timeperiod=n)
    # 12日量变动速率指标（Volume Rate of Change），以今天的成交量和N天前的成交量比较，
    # 通过计算某一段时间内成交量变动的幅度，应用成交量的移动比较来测量成交量运动趋向，
    # 达到事先探测成交量供需的强弱，进而分析成交量的发展趋势及其将来是否有转势的意愿，
    # 属于成交量的反趋向指标。属于成交量型因子
    for n in [6, 20]:
        df[f'VROCP{n}'] = talib.ROCP(volume_s, timeperiod=n)

    # RSI
    df['RSI'] = talib.RSI(close_s, timeperiod=14)

    # SAR 抛物线转向
    df['SAR'] = talib.SAR(high_s, low_s, acceleration=0.02, maximum=0.2)

    # TEMA
    for n in [6, 12, 26]:
        df[f'TEMA{n}'] = talib.TEMA(close_s, timeperiod=n)

    # TRANGE 真实范围
    df['TRANGE'] = talib.TRANGE(high_s, low_s, close_s)

    # TYPPRICE 典型价格
    df['TYPPRICE'] = talib.TYPPRICE(high_s, low_s, close_s)

    # TSF 时间序列预测
    df['TSF'] = talib.TSF(close_s, timeperiod=14)

    # ULTOSC 极限振子
    df['ULTOSC'] = talib.ULTOSC(high_s, low_s, close_s, timeperiod1=7, timeperiod2=14, timeperiod3=28)

    # 威廉指标
    df['WILLR'] = talib.WILLR(high_s, low_s, close_s, timeperiod=14)

    # 价格分位数水平
    data_list, data_count = [], [0]

    def get_index_pct(x):
        """获取当前价格在历史价格数组中的位置的百分比"""
        bisect.insort(data_list, x)
        data_count[0] += 1
        idx = bisect.bisect_left(data_list, x)
        return idx/data_count[0]

    df['index_pct'] = close_s.apply(get_index_pct)

    # 对 volume amount 取 log
    if log_av:
        df[volume_key] = np.log(volume_s.fillna(0) + 1)
        df[amount_key] = np.log(amount_s.fillna(0) + 1)

    if drop:
        df.dropna(inplace=True)

    return df


def get_factor(df: pd.DataFrame, trade_date_series=None, delivery_date_series=None,
               dropna=True, ohlcav_col_name_list=["open", "high", "low", "close", "amount", "volume"],
               do_multiple_factors=False) -> (pd.DataFrame, dict):
    """
    在当期时间序列数据基础上增加相关因子
    目前已经支持的因子包括量价因子、时间序列因子、交割日期因子
    :param df: 时间序列数据索引为日期
    :param trade_date_series: 交易日序列
    :param delivery_date_series:交割日序列
    :param dropna: 是否 dropna
    :param ohlcav_col_name_list: 量价因子相关列名称，默认["open", "high", "low", "close", "amount", "volume"]
    :param do_multiple_factors: 对数据进行倍增处理，将指定列乘以因子，如果 ！= None 则，返回dict{adj_factor: DataFrame}
    :return:
    """
    ret_df = df.copy()

    # 交易日相关因子
    if trade_date_series is not None:
        ret_df = add_factor_of_trade_date(ret_df, trade_date_series)

    # 交割日相关因子
    if delivery_date_series is not None:
        ret_df = add_factor_of_delivery_date(ret_df, delivery_date_series)

    # 增加量价因子
    train_df_dic = {}
    if ohlcav_col_name_list is not None:
        ret_df_tmp = ret_df.copy()
        ret_df = add_factor_of_price(ret_df, ohlcav_col_name_list)
        if do_multiple_factors:
            if dropna:
                ret_df.dropna(inplace=True)
            train_df_dic[1] = ret_df
            for adj_factor in [0.5, 0.75, 1.25, 1.5, 1.75, 2]:
                train_df_tmp = ret_df_tmp.copy()
                # 将 O,H,L,C,A 前五项进行因子扩充
                train_df_tmp.loc[:, ohlcav_col_name_list[:5]] *= adj_factor
                factor_df = add_factor_of_price(train_df_tmp, ohlcav_col_name_list=ohlcav_col_name_list)
                if dropna:
                    factor_df.dropna(inplace=True)
                train_df_dic[adj_factor] = factor_df

    if do_multiple_factors:
        return train_df_dic
    else:
        if dropna:
            ret_df.dropna(inplace=True)

        return ret_df


def _test_get_factor():
    from ibats_common.example.data import load_data, get_trade_date_series, get_delivery_date_series
    instrument_type = 'RU'
    file_name = f"{instrument_type}.csv"
    indexed_df = load_data(file_name).set_index('trade_date').drop('instrument_type', axis=1)
    indexed_df.index = pd.DatetimeIndex(indexed_df.index)
    # ohlcav_col_name_list is None
    factor_df = get_factor(indexed_df,
                           trade_date_series=get_trade_date_series(),
                           delivery_date_series=get_delivery_date_series(instrument_type))
    logger.info("data_multiplication_column_indexes=None\n%s\t%s", factor_df.shape, list(factor_df.columns))

    # ohlcav_col_name_list is not None
    ohlcav_col_name_list = ["open", "high", "low", "close", "amount", "volume"]
    train_df_dic = get_factor(indexed_df, ohlcav_col_name_list=ohlcav_col_name_list,
                              trade_date_series=get_trade_date_series(),
                              delivery_date_series=get_delivery_date_series(instrument_type))
    logger.info("ohlcav_col_name_list=%s", ohlcav_col_name_list)
    for adj_factor, train_df in train_df_dic.items():
        logger.info("adj_factor=%f, train_df: first close=%f\n%s\t%s",
                    adj_factor, train_df['close'].iloc[0], train_df.shape, list(train_df.columns))


def transfer_2_batch(df: pd.DataFrame, n_step, labels=None, date_from=None, date_to=None):
    """
    [num, factor_count] -> [num - n_step + 1, n_step, factor_count]
    将 df 转化成 n_step 长度的一段一段的数据
    labels 为与 df对应的数据，处理方式与index相同，如果labels不为空，则返回数据最后增加以下 new_ys
    :param df:
    :param n_step:
    :param labels:如果不为 None，则长度必须与 df.shape[0] 一致
    :param date_from:
    :param date_to:
    :return:
    """
    df_len = df.shape[0]
    if labels is not None and df_len != len(labels):
        raise ValueError("ys 长度 %d 必须与 df 长度 %d 保持一致", len(labels), df_len)
    # TODO: date_from, date_to 的逻辑可以进一步优化，延期为了省时间先保持这样
    # 根据 date_from 对factor进行截取
    if date_from is not None:
        date_from = pd.to_datetime(date_from)
        is_fit = df.index >= date_from
        if np.any(is_fit):
            start_idx = np.argmax(is_fit) - n_step

            if start_idx < 0:
                start_idx = 0
                logger.warning("%s 为起始日期的数据，前向历史数据不足 %d 条，因此，起始日期向后推移至 %s",
                               date_2_str(date_from), n_step, date_2_str(df.index[60]))

            df = df.iloc[start_idx:]
            df_len = df.shape[0]
            if labels is not None:
                labels = labels[start_idx:]

        else:
            logger.warning("没有 %s 之后的数据，当前数据最晚日期为 %s",
                           date_2_str(date_from), date_2_str(max(df.index)))
            if labels is not None:
                return None, None, None, None
            else:
                return None, None, None

    # 根据 date_from 对factor进行截取
    if date_to is not None:
        date_to = pd.to_datetime(date_to)
        is_fit = df.index <= date_to
        if np.any(is_fit):
            to_idx = np.argmin(is_fit)
            df = df.iloc[:to_idx]
            df_len = df.shape[0]
            if labels is not None:
                labels = labels[:to_idx]

        else:
            logger.warning("没有 %s 之前的数据，当前数据最晚日期为 %s",
                           date_2_str(date_to), date_2_str(min(df.index)))
            if labels is not None:
                return None, None, None, None
            else:
                return None, None, None

    new_shape = [df_len - n_step + 1, n_step]
    new_shape.extend(df.shape[1:])
    df_index, df_columns = df.index[(n_step - 1):], df.columns
    data_arr_batch, factor_arr = np.zeros(new_shape), df.to_numpy()

    for idx_from, idx_to in enumerate(range(n_step, factor_arr.shape[0] + 1)):
        data_arr_batch[idx_from] = factor_arr[idx_from: idx_to]

    if labels is not None:
        new_ys = labels[(n_step - 1):]
        return df_index, df_columns, data_arr_batch, new_ys
    else:
        return df_index, df_columns, data_arr_batch


def _test_get_batch_factor():
    data_len = 8
    date_arr = pd.date_range(pd.to_datetime('2018-01-01'),
                             pd.to_datetime('2018-01-01') + pd.Timedelta(days=data_len * 2 - 1),
                             freq=pd.Timedelta(days=2))
    date_index = pd.DatetimeIndex(date_arr)
    df = pd.DataFrame(
        {'a': list(range(data_len)),
         'b': list(range(data_len*2, data_len * 3)),
         'c': list(range(data_len * 10, data_len * 11))},
        index=date_index,
    )
    labels = list(range(data_len))
    print("df\n", df)
    n_step = 5
    df_index, df_columns, data_arr_batch, new_labels = transfer_2_batch(df, n_step, labels)
    print("new df_index", df_index)
    print("new factor_columns", df_columns)
    print('new reshaped data_arr_batch')
    print(data_arr_batch)
    print("df.shape: ", df.shape)
    print("new_shape:", data_arr_batch.shape)
    print("new_labels:", new_labels)

    date_from = '2018-01-13'
    df_index, df_columns, data_arr_batch, new_labels = transfer_2_batch(df, n_step, labels, date_from=date_from)
    print('date_from=', date_from)
    print("new df_index", df_index)
    print("new factor_columns", df_columns)
    print('new reshaped data_arr_batch')
    print(data_arr_batch)
    print("df.shape: ", df.shape)
    print("new_shape:", data_arr_batch.shape)
    print("new_labels:", new_labels)


def get_batch_factor(md_df: pd.DataFrame, n_step, labels=None, **factor_kwargs):
    """
    get_factor(...) -> transfer_2_batch(...)
    :param md_df:
    :param n_step:
    :param labels:
    :param factor_kwargs:
    :return:
    """
    factor_df = get_factor(md_df, **factor_kwargs)
    df_index, df_columns, data_arr_batch = transfer_2_batch(factor_df, n_step=n_step, labels=labels)
    return df_index, df_columns, data_arr_batch


if __name__ == '__main__':
    # _test_get_factor()
    _test_get_batch_factor()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/4/7 16:31
@File    : plot.py
@contact : mmmaaaggg@163.com
@desc    :
2020-01-22 remove " and is_windows_os()" on if condition of plot_or_show function
"""
import itertools
import logging
import os
from datetime import datetime
from functools import lru_cache

import ffn
import matplotlib
# matplotlib.use('Qt5Agg')  # 需要 pip3 install PyQt5，windows 下無效
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from ibats_utils.mess import date_2_str, is_windows_os, open_file_with_system_app
from matplotlib.cm import get_cmap
from matplotlib.font_manager import FontProperties
from pandas.plotting import register_matplotlib_converters
from scipy import stats

from ibats_common.backend.mess import get_cache_folder_path
from ibats_common.backend.label import calc_label2, calc_label3

logger = logging.getLogger(__name__)
logger.debug("matplotlib.backend => %s", matplotlib.get_backend())
register_matplotlib_converters()
FFN_VERSION = ffn.__version__
ALTER_BG_COLOR = '#e0e0e0'


def get_file_name(header, name=None):
    file_name = f'{header} {np.random.randint(10000)}.png' if name is None else f'{header} {name}.png'
    return file_name


def clean_cache():
    """
    清空cache目录
    :return:
    """
    folder_path = get_cache_folder_path()
    for root, dirs, files in os.walk(folder_path, topdown=False):
        # print(root, dirs, files)
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))


def drawdown_plot(df: pd.DataFrame, perf_stats=None, col_name_list=None,
                  enable_show_plot=True, enable_save_plot=False, name=None, stg_run_id=None):
    """
    回撤曲线
    :param df:
    :param perf_stats:
    :param col_name_list:
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :param stg_run_id:
    :return:
    """
    if col_name_list is None:
        data_df = df.to_drawdown_series()
        col_name_list = list(df.columns)
    else:
        data_df = df[col_name_list].to_drawdown_series()

    if perf_stats is None:
        perf_stats = df.calc_stats()

    col_mdd_len_dic = {col_name: s.drawdown_details.Length.max()
                       for col_name, s in perf_stats.items() if s.drawdown_details is not None}
    col_mdd_dic = {col_name: mdd for col_name, mdd in data_df.min().items()}
    data_df.rename(
        columns={col_name: f"{col_name}[{mdd * 100:.2f}% {col_mdd_len_dic.setdefault(col_name, '')}]"
                 for col_name, mdd in col_mdd_dic.items()},
        inplace=True)

    ax = data_df.plot()
    ax.set_title(f"Drawdown {['{:.2f}%'.format(col_mdd_dic[_] * 100) for _ in col_name_list]}")

    file_name = get_file_name(f'drawdown', name=name)
    file_path = plot_or_show(enable_save_plot=enable_save_plot, enable_show_plot=enable_show_plot,
                             file_name=file_name, stg_run_id=stg_run_id)

    return data_df, file_path


def plot_corr(df: pd.DataFrame, perf_stats=None,
              enable_show_plot=True, enable_save_plot=False, name=None, stg_run_id=None):
    """
    相关性矩阵图
    :param df:
    :param perf_stats:
    :param col_name_list:
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :param stg_run_id:
    :return:
    """
    if perf_stats is None:
        perf_stats = df.calc_stats()

    ax = perf_stats.plot_correlation()
    plt.suptitle("Correlation")

    file_name = get_file_name(f'correlation', name=name)
    file_path = plot_or_show(enable_save_plot=enable_save_plot, enable_show_plot=enable_show_plot,
                             file_name=file_name, stg_run_id=stg_run_id)
    return file_path


def plot_rr_df(df: pd.DataFrame, col_name_list=None, enable_show_plot=True, enable_save_plot=False,
               name=None, stg_run_id=None):
    """
    Return Rate
    :param df:
    :param col_name_list:
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :param stg_run_id:
    :return:
    """
    if col_name_list is not None:
        data_df = df[col_name_list]
    else:
        data_df = df

    # """绘图函数"""
    ax = data_df.plot(grid=True)
    ax.set_title(
        f"Return Rate " if name is None else f"Return Rate [{name}] "
        f"{date_2_str(min(data_df.index))} - {date_2_str(max(data_df.index))} ({data_df.shape[0]} days)")

    file_name = get_file_name(f'rr', name=name)
    file_path = plot_or_show(enable_save_plot=enable_save_plot, enable_show_plot=enable_show_plot,
                             file_name=file_name, stg_run_id=stg_run_id)

    return file_path


def _test_plot_rr_df():
    from ibats_common.example.data import load_data
    df = load_data('RB.csv').set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    file_path = plot_rr_df(df, col_name_list=['close'],
                           enable_show_plot=True, enable_save_plot=True, name='test_use', stg_run_id=2)
    logger.info(file_path)


def plot_scatter_matrix(df: pd.DataFrame, diagonal='hist', col_name_list=None, enable_show_plot=True,
                        enable_save_plot=False, name=None, stg_run_id=None):
    """
    plot scatter_matrix
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.plotting.scatter_matrix.html
    :param df:
    :param diagonal: diagonal，必须且只能在{'hist', 'kde'}中选择1个，
    'hist'表示直方图(Histogram plot),'kde'表示核密度估计(Kernel Density Estimation)
    该参数是scatter_matrix函数的关键参数
    :param col_name_list:
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :param stg_run_id:
    :return:
    """
    if col_name_list is not None:
        data_df = df[col_name_list]
    else:
        data_df = df

    # """绘图函数"""
    pd.plotting.scatter_matrix(data_df)
    plt.suptitle(
        f"Scatter Matrix " if name is None else f"Scatter Matrix [{name}] "
        f"{date_2_str(min(data_df.index))} - {date_2_str(max(data_df.index))} ({data_df.shape[0]} days)")

    file_name = get_file_name('scatter_matrix', name=name)
    file_path = plot_or_show(enable_save_plot=enable_save_plot, enable_show_plot=enable_show_plot,
                             file_name=file_name, stg_run_id=stg_run_id)
    return file_path


def hist_norm(data, bins=10, enable_show_plot=True, enable_save_plot=False, name=None, stg_run_id=None):
    """
    hist 分布图及正太分布曲线
    :param data:
    :param bins:
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :param stg_run_id:
    :return:
    """
    fig, ax = plt.subplots()
    # the histogram of the data
    _, _, patches = ax.hist(data, bins, density=True)
    n, bins_v, mean, std = plot_norm(data, bins=bins, ax=ax)
    # ax.set_xlabel('pct change')
    # ax.set_ylabel('change rate')
    ax.set_title(f"{'Data' if name is None else name} Histogram (mean={mean:.4f} std={std:.4f})")

    file_name = get_file_name(f'hist', name=name)
    rr_plot_file_path = plot_or_show(enable_show_plot=enable_show_plot, enable_save_plot=enable_save_plot,
                                     file_name=file_name, stg_run_id=stg_run_id)
    return n, bins_v, rr_plot_file_path


def plot_norm(data: pd.Series, bins=10, ax=None, is_show_plot=None):
    """
    显示当前数据的正太分布曲线
    :param data:
    :param bins: bar 数量
    :param ax: 如果为None，则新建一个画布
    :param is_show_plot: 是否展示
    :return: n, bins_v, mean, std
    """
    if ax is None:
        fig, ax = plt.subplots()
        if is_show_plot is None:
            is_show_plot = True

    if is_show_plot is None:
        is_show_plot = False

    n, bins_v = np.histogram(data, bins=bins)

    mu = data.mean()  # mean of distribution
    sigma = data.std()  # standard deviation of distribution
    # def norm_func(x, mu, sigma):
    #     pdf = np.exp(-((x - mu)**2)/(2*sigma**2)) / (sigma * np.sqrt(2*np.pi))
    #     return pdf
    # y = norm_func(bins, mu, sigma)  # 与 mlab.normpdf(bins, mu, sigma) 相同
    # y = mlab.normpdf(bins, mu, sigma)
    y = stats.norm.pdf(bins_v, loc=mu, scale=sigma)
    ax.plot(bins_v, y, '--')
    plt.grid(True)
    if is_show_plot:
        plt.show()

    return n, bins_v, mu, sigma


def hist_norm_sns(data, bins=10):
    """sns 方式 hist 分布图及正太分布曲线"""
    sns.set_palette("hls")
    sns.distplot(
        data, bins=bins,
        fit=stats.norm,
        kde_kws={"color": "darkorange", "lw": 1, "label": "KDE", "linestyle": "--"},
        hist_kws={"color": "darkblue"})
    plt.grid(True)
    plt.show()


def wave_hist(df: pd.DataFrame, columns=None, bins=50, figure_4_each_col=True,
              col_transfer_dic: (dict, None) = None, enable_show_plot=True, enable_save_plot=False,
              name=None, stg_run_id=None):
    """
    波动率分布图
    :param df:
    :param perf_stats: df.calc_stats()
    :param columns: 显示哪些列
    :param bins: bar 数量
    :param figure_4_each_col: 每一个col单独一张图
    :param col_transfer_dic: 列值转换
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :param stg_run_id:
    :return:
    """
    if columns is not None:
        data_df = df.loc[:, [_ for _ in columns if _ in df.columns]].copy()
    else:
        data_df = df.copy()

    data_df.dropna(inplace=True)
    if col_transfer_dic is not None:
        for func, col_name_list in col_transfer_dic.items():
            if func == 'return':
                rename_dic = {}
                for col_name in col_transfer_dic[func]:
                    if col_name not in data_df:
                        continue
                    rename_dic[col_name] = f"{col_name} {func}"
                    data_df.loc[:, col_name] = data_df[col_name].to_returns()
                else:
                    data_df.rename(columns=rename_dic, inplace=True)

            elif func == 'pct_change':
                rename_dic = {}
                for col_name in col_transfer_dic[func]:
                    if col_name not in data_df:
                        continue
                    rename_dic[col_name] = f"{col_name} {func}"
                    data_df.loc[:, col_name] = data_df[col_name].pct_change()
                else:
                    data_df.rename(columns=rename_dic, inplace=True)
            else:
                rename_dic = {}
                for col_name in col_transfer_dic[func]:
                    if col_name not in data_df:
                        continue
                    rename_dic[col_name] = f"{col_name} {str(func)}"
                    data_df.loc[:, col_name] = func(data_df[col_name])
                else:
                    data_df.rename(columns=rename_dic, inplace=True)

        data_df.dropna(inplace=True)

    columns = list(data_df.columns)
    n_bins_dic, file_path = {}, None
    if figure_4_each_col:
        for col_name in columns:
            data = data_df[col_name]
            try:
                data = data[~np.isinf(data)].dropna()
            except TypeError as exp:
                logger.exception('column %s 数据类型无效，无法进行 np.isinf 计算', col_name)
                raise exp from exp

            n, bins_v, file_path_sub = hist_norm(
                data, bins=bins, enable_save_plot=enable_save_plot, enable_show_plot=enable_show_plot,
                name=col_name, stg_run_id=stg_run_id)
            n_bins_dic[col_name] = (n, bins_v)
            if file_path is None:
                file_path = [file_path_sub]
            else:
                file_path.append(file_path_sub)

    else:

        ax = data_df.hist(bins=bins)
        if isinstance(ax, np.ndarray):
            # ax 总是一个 偶数 出现，因此，需要进行一次长度对齐
            ax_list = list(itertools.chain(*ax))[:len(columns)]
        else:
            ax_list = [ax]

        for col_name, ax_sub in zip(columns, ax_list):
            data = data_df[col_name]
            try:
                data = data[~np.isinf(data)].dropna()
            except TypeError as exp:
                logger.exception('column %s 数据类型无效，无法进行 np.isinf 计算', col_name)
                raise exp from exp

            n, bins_v, mean, std = plot_norm(data, bins=bins, ax=ax_sub)
            # 字太多无法显示
            # ax_sub.set_title(f"{col_name}\n(mean={mean:.4f} std={std:.4f})")
            n_bins_dic[col_name] = (n, bins_v)

        plt.grid(True)
        file_name = get_file_name(f'hist', name=name)
        file_path = plot_or_show(enable_show_plot=enable_show_plot, enable_save_plot=enable_save_plot,
                                 file_name=file_name, stg_run_id=stg_run_id)

    return n_bins_dic, file_path


def _test_wave_hist():
    from ibats_common.example.data import load_data
    df = load_data('RB.csv').set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    col_transfer_dic = {
        'return': ['open', 'high', 'low', 'close', 'volume']
    }
    n_bins_dic, file_path = wave_hist(df, figure_4_each_col=False, col_transfer_dic=col_transfer_dic,
                                      enable_save_plot=True, enable_show_plot=True, name="test_use", stg_run_id=1)


def hist_n_rr(df: pd.DataFrame, n_days, columns=None, bins=50,
              enable_show_plot=True, enable_save_plot=False, name=None, stg_run_id=None):
    """
    未来N日收益率分布图
    :param df:
    :param n_days: 计算未来 N 日的收益率最高值，最低值
    :param columns: 显示哪些列
    :param bins: bar 数量
    :param enable_show_plot:
    :param enable_save_plot:
    :param name:
    :return:
    """
    if columns is not None:
        data_df = df.loc[:, columns].dropna().copy()
    else:
        data_df = df.dropna().copy()

    column_name_list, df_dic, quantile_dic, n_bins_dic, file_path_dic = list(data_df.columns), {}, {}, {}, {}
    for n_day in n_days:
        # 计算各个列未来N日收益率波动
        max_df = data_df.rename(
            columns={_: _ + ' max rr' for _ in column_name_list}
        ).rolling(n_day).apply(
            lambda x: max(x / x[0]) - 1, raw=True).shift(-(n_day - 1))
        min_df = data_df.rename(
            columns={_: _ + ' min rr' for _ in column_name_list}
        ).rolling(n_day).apply(
            lambda x: min(x / x[0]) - 1, raw=True).shift(-(n_day - 1))
        # 合并数据
        merged_df = max_df.join(min_df).dropna()
        for col_name in column_name_list:
            col_names = [col_name + ' max rr', col_name + ' min rr']
            new_df = merged_df[col_names].copy()
            df_list = [new_df]
            # 计算收益率分位数信息
            idxmax_s = abs(new_df).idxmax(axis=1)
            # df.loc[idxmax_s != 'close max rr', 'close max rr'] = np.nan
            # df.loc[idxmax_s != 'close min rr', 'close min rr'] = np.nan
            label_quantile_dic = {}
            for label in new_df.columns:
                data_s = new_df.loc[idxmax_s == label, label].dropna()
                df_list.append(data_s)
                label_quantile_dic[label] = data_s.quantile([0.2, 0.33, 0.5, 0.66, 0.8])

            df_dic[(n_day, col_name)] = df_list
            quantile_dic[(n_day, col_name)] = pd.DataFrame(label_quantile_dic).T

    # 输出图片
    for (n_day, col_name), df_list in df_dic.items():
        hist_kwargs = dict(histtype='stepfilled', alpha=0.8, bins=bins)
        n_list, bins_v_list, axis_list, labels = [], [], [], []
        for data_s in df_list[1:]:
            n, bins_v, axis = plt.hist(data_s.dropna(), **hist_kwargs)
            labels.append(data_s.name)
            axis_list.extend(axis)

        plt.grid(True)
        plt.legend(axis_list, labels, loc='upper right')
        if name is None:
            title = f'hist of max/min {col_name} rr in {n_day} days'
        else:
            title = f'{name} hist of max/min {col_name} rr in {n_day} days'

        plt.suptitle(title)
        n_bins_dic[(n_day, col_name)] = (n_list, bins_v_list)

        file_name = get_file_name(f'hist {n_day} {col_name} future', name=name)
        file_path = plot_or_show(enable_save_plot=enable_save_plot, enable_show_plot=enable_show_plot,
                                 file_name=file_name, stg_run_id=stg_run_id)
        if file_path is not None:
            file_path_dic[(n_day, col_name)] = file_path

    ret_dic = dict(df_dic=df_dic, n_bins_dic=n_bins_dic, quantile_dic=quantile_dic)
    return ret_dic, file_path_dic


def _test_hist_n_rr():
    from ibats_common.example.data import load_data
    df = load_data('RB.csv').set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    ret_dic, file_path_dic = hist_n_rr(df, n_days=[3, 5], columns=['close'], enable_save_plot=True, stg_run_id=3)
    for k, v in ret_dic['quantile_dic'].items():
        logger.info('%s -> \n%s', k, v)


def label_distribution(close_df: pd.DataFrame, min_rr: float, max_rr: float, max_future: int,
                       name=None, **enable_kwargs):
    """
    输出分类标签在行情图中的分布情况
    :param close_df:
    :param min_rr:
    :param max_rr:
    :param max_future:为空则进行2分类，不为空则3分类
    :param enable_kwargs:
    :param name:
    :return:
    """
    logger.debug('%s [%f ~ %f] max_future=%d, name="%s"', close_df.shape, min_rr, max_rr, max_future, name)
    value_arr = close_df.to_numpy()
    if max_future is None:
        target_arr = calc_label2(value_arr, min_rr, max_rr, one_hot=False, dtype='int')
    else:
        target_arr = calc_label3(value_arr, min_rr, max_rr, max_future=max_future, one_hot=False, dtype='int')

    ax = close_df.plot(grid=True)
    plt.suptitle(f'label [{min_rr * 100:.2f}% ~ {max_rr * 100:.2f}%]')
    x_values = list(close_df.index)
    colors = [None, ALTER_BG_COLOR, '#d62728']

    value_count = len(value_arr)
    label_list = list(set(target_arr))
    distribution_rate_df = pd.DataFrame(
        {_: {'pct': sum(target_arr == _) / value_count} for _ in label_list}, columns=label_list)
    for label, range_from, range_to in get_range_num_iter(target_arr):
        color = colors[int(label)]
        if color is None:
            continue
        p = plt.axvspan(x_values[range_from], x_values[range_to], facecolor=color, alpha=0.5)

    file_name = get_file_name(f'label distribution', name=name)
    file_path = plot_or_show(file_name=file_name, **enable_kwargs)
    return distribution_rate_df, file_path


def show_dl_accuracy(real_ys, pred_ys, close_df: pd.DataFrame, split_point_list=None,
                     base_line_list: (None, dict) = None, show_moving_avg=False):
    """
    将实际标记与预测标记进行对比并结合行情close_df显示在plot图上
    :param real_ys:
    :param pred_ys:
    :param close_df:
    :param split_point_list:
    :param base_line_list:
    :param show_moving_avg: 展示滚动平均收益率，如果展示，则会显示 三张图，否则两张
    :return:
    """
    trade_date_index = close_df.index
    date_from_str, date_to_str = date_2_str(trade_date_index[0]), date_2_str(trade_date_index[-1])
    # 检查长度是否一致
    if not (len(real_ys) == len(pred_ys) == close_df.shape[0]) or len(real_ys) == 0:
        logger.error("[%s - %s] len(real_ys)=%d, len(pred_ys)=%d, close_df.shape[0]=%d 不一致",
                     date_from_str, date_to_str, len(real_ys), len(pred_ys), close_df.shape[0])
        return
    enable_kwargs = dict(enable_save_plot=False, enable_show_plot=False, do_clr=False)
    if show_moving_avg:
        fig = plt.figure(figsize=(12, 16))
        fig_idx = 310
    else:
        fig = plt.figure(figsize=(12, 14))
        fig_idx = 210

    # 分析成功率
    # 第一章图：累计平均成功率
    is_fit_arr = pred_ys == real_ys
    accuracy = sum(is_fit_arr) / len(pred_ys) * 100
    logger.info("模型准确率 [%s - %s] accuracy: %.2f%%", date_from_str, date_to_str, accuracy)
    accuracy_list, fit_sum = [], 0
    for tot_count, (is_fit, trade_date) in enumerate(zip(is_fit_arr, trade_date_index), start=1):
        if is_fit:
            fit_sum += 1
        accuracy_list.append(fit_sum / tot_count)

    accuracy_df = pd.DataFrame({'accuracy': accuracy_list}, index=trade_date_index)
    ax = fig.add_subplot(fig_idx + 1)
    plot_accuracy(accuracy_df, close_df, ax=ax, base_line_list=base_line_list,
                  name=f'Accumulation Avg Accuracy [{date_from_str}{date_to_str}]',
                  split_point_list=split_point_list, **enable_kwargs)

    # 第一章图：累计平均成功率
    real_pred_df = pd.DataFrame({'pred': pred_ys, 'real': real_ys}, index=trade_date_index)
    ax2 = fig.add_subplot(fig_idx + 2)
    plot_accuracy(accuracy_df, close_df, ax=ax2, base_line_list=base_line_list,
                  name=f'Accumulation Avg Accuracy [{date_from_str}{date_to_str}]',
                  real_pred_df=real_pred_df, **enable_kwargs)
    # 第三章图：移动平均成功率
    if show_moving_avg:
        accuracy_list, win_size = [], 30
        is_fit_df = pd.DataFrame({'fit': is_fit_arr}, index=trade_date_index)
        accuracy_df = is_fit_df.rolling(win_size, min_periods=1).apply(lambda x: sum(x) / len(x), raw=True)
        for idx in range(win_size, len(is_fit_arr)):
            accuracy_list.append(sum(is_fit_arr[idx - win_size:idx] / win_size))

        # close2_df = close_df.iloc[win_size:]
        # accuracy_df = pd.DataFrame({'accuracy': accuracy_list}, index=close2_df.index)
        ax3 = fig.add_subplot(fig_idx + 3)
        plot_accuracy(accuracy_df, close_df, ax=ax3, base_line_list=base_line_list,
                      name=f'{win_size} Moving Avg Accuracy [{date_from_str}{date_to_str}]',
                      split_point_list=split_point_list, **enable_kwargs)

    # 展示图片
    file_name = f"accuracy [{date_from_str}-{date_to_str}].png"
    file_path = plot_or_show(enable_save_plot=True, enable_show_plot=True, file_name=file_name)
    return file_path


def _test_show_dl_accuracy():
    real_ys, pred_ys = np.random.randint(1, 3, size=100), np.random.randint(1, 3, size=100)
    date_arr = pd.date_range(pd.to_datetime('2018-01-01'),
                             pd.to_datetime('2018-01-01') + pd.Timedelta(days=99))
    date_index = pd.DatetimeIndex(date_arr)
    close_df = pd.DataFrame({'close': np.sin(np.linspace(0, 10, 100))}, index=date_index)

    split_point_list = np.random.randint(len(date_arr), size=10)
    split_point_list.sort()
    split_point_list = date_arr[split_point_list]
    base_line_list = [0.3, 0.6]

    show_dl_accuracy(real_ys, pred_ys, close_df, split_point_list, base_line_list, show_moving_avg=True)


def show_drl_accuracy(real_label_s, action_s, close_df: pd.DataFrame, split_point_list=None,
                      show_moving_avg=False):
    """
    将实际标记与预测标记进行对比并结合行情close_df显示在plot图上
    :param real_label_s:
    :param action_s:
    :param close_df:
    :param split_point_list:
    :param show_moving_avg: 展示滚动平均收益率，如果展示，则会显示 三张图，否则两张
    :return:
    """
    trade_date_index = close_df.index
    date_from_str, date_to_str = date_2_str(trade_date_index[0]), date_2_str(trade_date_index[-1])
    # 检查长度是否一致
    if not (len(real_label_s) == len(action_s) == close_df.shape[0]) or len(real_label_s) == 0:
        logger.error("[%s - %s] len(real_label_s)=%d, len(action_s)=%d, close_df.shape[0]=%d 不一致",
                     date_from_str, date_to_str, len(real_label_s), len(action_s), close_df.shape[0])
        return
    enable_kwargs = dict(enable_save_plot=False, enable_show_plot=False, do_clr=False)
    if show_moving_avg:
        fig = plt.figure(figsize=(12, 16))
        fig_idx = 310
    else:
        fig = plt.figure(figsize=(12, 14))
        fig_idx = 210

    # 分析成功率
    # 第一章图：累计平均成功率
    is_fit_arr = action_s == real_label_s
    accuracy = sum(is_fit_arr) / len(action_s) * 100
    logger.info("模型准确率 [%s - %s] accuracy: %.2f%%", date_from_str, date_to_str, accuracy)
    accuracy_list, fit_sum = [], 0
    for tot_count, (is_fit, trade_date) in enumerate(zip(is_fit_arr, trade_date_index), start=1):
        if is_fit:
            fit_sum += 1
        accuracy_list.append(fit_sum / tot_count)

    accuracy_df = pd.DataFrame({'accuracy': accuracy_list}, index=trade_date_index)
    ax = fig.add_subplot(fig_idx + 1)
    plot_accuracy(accuracy_df, close_df, ax=ax,
                  name=f'Accumulation Avg Accuracy [{date_from_str}{date_to_str}]',
                  split_point_list=split_point_list, **enable_kwargs)

    # 第一章图：累计平均成功率
    real_pred_df = pd.DataFrame({'pred': action_s, 'real': real_label_s}, index=trade_date_index)
    ax2 = fig.add_subplot(fig_idx + 2)
    plot_accuracy(accuracy_df, close_df, ax=ax2,
                  name=f'Accumulation Avg Accuracy [{date_from_str}{date_to_str}]',
                  real_pred_df=real_pred_df, **enable_kwargs)
    # 第三章图：移动平均成功率
    if show_moving_avg:
        accuracy_list, win_size = [], 30
        is_fit_df = pd.DataFrame({'fit': is_fit_arr}, index=trade_date_index)
        accuracy_df = is_fit_df.rolling(win_size, min_periods=1).apply(lambda x: sum(x) / len(x), raw=True)
        for idx in range(win_size, len(is_fit_arr)):
            accuracy_list.append(sum(is_fit_arr[idx - win_size:idx] / win_size))

        # close2_df = close_df.iloc[win_size:]
        # accuracy_df = pd.DataFrame({'accuracy': accuracy_list}, index=close2_df.index)
        ax3 = fig.add_subplot(fig_idx + 3)
        plot_accuracy(accuracy_df, close_df, ax=ax3,
                      name=f'{win_size} Moving Avg Accuracy [{date_from_str}{date_to_str}]',
                      split_point_list=split_point_list, **enable_kwargs)

    # 展示图片
    file_name = f"accuracy [{date_from_str}-{date_to_str}].png"
    file_path = plot_or_show(enable_save_plot=True, enable_show_plot=True, file_name=file_name)
    return file_path


def plot_or_show(enable_save_plot=True, enable_show_plot=True, file_name=None, stg_run_id=None, do_clr=True,
                 folder_path=None):
    if enable_save_plot:
        if folder_path is None:
            if stg_run_id is None:
                folder_path = get_cache_folder_path()
            else:
                folder_path = os.path.join(get_cache_folder_path(), str(stg_run_id))
        else:
            folder_path = os.path.abspath(folder_path)

        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        if file_name is None:
            file_name = 'output.png'
        if not file_name.endswith('.png'):
            file_name += '.png'
        file_path = os.path.join(folder_path, file_name)
        logger.debug("save to %s", file_path)
        plt.savefig(file_path, dpi=75)
    else:
        file_path = None

    if enable_show_plot:
        # 2020-01-22 remove " and is_windows_os()" on if condition of plot_or_show function
        if file_path is not None:
            open_file_with_system_app(file_path)
        else:
            plt.show()

    if do_clr:
        plt.cla()
        plt.clf()
        plt.close()

    return file_path


def _test_label_distribution():
    from ibats_common.example.data import load_data
    df = load_data('RB.csv').set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    distribution_rate_df, file_path = label_distribution(df['close'], min_rr=-0.01, max_rr=0.01, max_future=3,
                                                         enable_save_plot=True, enable_show_plot=True, stg_run_id=4)
    logger.info('\n%s', distribution_rate_df)
    logger.info(file_path)


def get_range_num_iter(arr):
    val_cur, range_from, range_to = arr[0], 0, None
    for range_to, val in enumerate(arr[1:], start=1):
        if val != val_cur:
            yield val_cur, range_from, range_to
            range_from = range_to
            val_cur = val
    else:
        range_to += 1
        yield val_cur, range_from, range_to


def _test_n_days_rr_distribution():
    from ibats_common.example.data import load_data
    df = load_data('RB.csv').set_index('trade_date').drop('instrument_type', axis=1)
    df.index = pd.DatetimeIndex(df.index)
    ret_dic, file_path_dic, enable_kwargs = {}, {}, dict(enable_save_plot=True, enable_show_plot=True, stg_run_id=5)
    tmp_dic, file_path = hist_n_rr(df, n_days=[3, 5], columns=['close'], **enable_kwargs)
    ret_dic['hist_future_n_rr'] = tmp_dic
    file_path_dic['hist_future_n_rr'] = file_path

    from collections import defaultdict
    file_path_dic['label_distribution'] = defaultdict(dict)
    ret_dic['label_distribution'] = defaultdict(dict)
    for (n_day, col_name), quantile_df in tmp_dic['quantile_dic'].items():
        path_dic = file_path_dic['label_distribution'][(n_day, col_name)]
        distribution_dic = ret_dic['label_distribution'][(n_day, col_name)]
        col_count = quantile_df.shape[1]
        for n in range(col_count):
            max_rr = quantile_df.iloc[0, n]
            min_rr = quantile_df.iloc[1, col_count - n - 1]
            distribution_rate_df, file_path = label_distribution(
                df['close'], min_rr=min_rr, max_rr=max_rr, max_future=n_day,
                name=f"{col_name}[{min_rr * 100:.2f}%-{max_rr * 100:.2f}%]", **enable_kwargs)
            path_dic[(min_rr, max_rr)] = file_path
            distribution_dic[(min_rr, max_rr)] = distribution_rate_df


def plot_accuracy(accuracy_df, close_df, split_point_list=None, ax=None,
                  name=None, base_line_list: (None, dict) = None, real_pred_df=None, **enable_kwargs):
    """
    显示成功率与行情叠加走势，如果 real_pred_df is not None 则叠加预测与真实标记
    :param accuracy_df:
    :param close_df:
    :param split_point_list:
    :param ax:
    :param name:
    :param base_line_list:
    :param real_pred_df:
    :param enable_kwargs:
    :return:
    """
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    accuracy_df = accuracy_df.copy()
    l1 = ax.plot(accuracy_df, color='r', label='accuracy')
    if real_pred_df is not None:
        for x0, x1, data in get_data_range_iter(real_pred_df['pred'], extent_left=True):
            if data == 0:
                p = plt.axvspan(x0, x1, ymin=0.66, ymax=1, facecolor='cornflowerblue', alpha=0.5)
            elif data == 1:
                p = plt.axvspan(x0, x1, ymin=0.66, ymax=1, facecolor='lightcoral', alpha=0.5)
            elif data == 2:
                p = plt.axvspan(x0, x1, ymin=0.66, ymax=1, facecolor='lightseagreen', alpha=0.5)

        for x0, x1, data in get_data_range_iter(real_pred_df['real'], extent_left=True):
            if data == 0:
                p = plt.axvspan(x0, x1, ymin=0, ymax=0.33, facecolor='cornflowerblue', alpha=0.5)
            elif data == 1:
                p = plt.axvspan(x0, x1, ymin=0, ymax=0.33, facecolor='lightcoral', alpha=0.5)
            elif data == 2:
                p = plt.axvspan(x0, x1, ymin=0, ymax=0.33, facecolor='lightseagreen', alpha=0.5)

        is_fit_s = real_pred_df.apply(lambda x: x['real'] if x['real'] == x['pred'] else -2, axis=1)
        for x0, x1, data in get_data_range_iter(is_fit_s, extent_left=True):
            if data == 0:
                p = plt.axvspan(x0, x1, ymin=0.33, ymax=0.66, facecolor='cornflowerblue', alpha=0.5)
            elif data == 1:
                p = plt.axvspan(x0, x1, ymin=0.33, ymax=0.66, facecolor='lightcoral', alpha=0.5)
            elif data == 2:
                p = plt.axvspan(x0, x1, ymin=0.33, ymax=0.66, facecolor='lightseagreen', alpha=0.5)
            else:
                p = plt.axvspan(x0, x1, ymin=0.33, ymax=0.66, facecolor='red', alpha=0.5)

    ax2 = ax.twinx()
    l2 = ax2.plot(close_df, label='md')
    lns = l1 + l2
    if base_line_list is not None:
        acc = base_line_list[0]
        if acc is not None and acc != 0:
            accuracy_df['train_accuracy'] = acc
            lns += ax.plot(accuracy_df['train_accuracy'], linestyle='--', color='b', label='train_accuracy')
        acc = base_line_list[1]
        if acc is not None and acc != 0:
            accuracy_df['validation_accuracy'] = acc
            lns += ax.plot(accuracy_df['validation_accuracy'], linestyle='--', color='r', label='validation_accuracy')

    accuracy_df['50%'] = 0.5
    lns += ax.plot(accuracy_df['50%'], color='b', label='train_accuracy')

    plt.legend(lns, [_.get_label() for _ in lns], loc=0)
    plt.grid(True)
    if name is not None:
        # 解决中文字符无法显示问题，稍后将其范化
        # font = get_font_properties()
        # plt.suptitle(name, fontproperties=font)
        plt.title(name)

    # 分段着色
    if split_point_list is not None and len(split_point_list) > 2:
        x0, x1 = None, None
        for num, point in enumerate(split_point_list):
            if num % 2 == 0:
                x0 = point
            else:
                x1 = point
                if num >= 1:
                    p = plt.axvspan(x0, x1, facecolor=ALTER_BG_COLOR, alpha=0.5)

    datetime_str = datetime.now().strftime('%Y-%m-%d_%H_%M_%S_%f')
    file_name = f"{datetime_str}" if name is None else f"{name}_{datetime_str}"
    file_path = plot_or_show(file_name=file_name, **enable_kwargs)
    return file_path


def plot_twin(df_list, df2, ax=None, name=None, enable_save_plot=True, enable_show_plot=True, do_clr=True,
              folder_path=None, y_scales_log=[False, False], in_sample_date_line=None, figsize=(6, 8)):
    """输出双坐标中图像"""
    if ax is None:
        fig = plt.figure(figsize=figsize)  #
        ax = fig.add_subplot(111)

    ax.set_prop_cycle(color=get_cmap('tab20').colors)
    if not isinstance(df_list, list):
        df_list = [df_list]

    l1, legend1, min_x, max_x = [], [], None, None
    for df, linestyle in zip(df_list,
                             ['-', ':', 'dashdotdotted', 'densely dashdotdotted', 'densely dotted'][:len(df_list)]):
        if df.shape[0] == 0:
            continue
        # 绘图
        if isinstance(df, pd.DataFrame):
            if df.shape[1] == 1:
                legend1 += [df.columns[0]]
            else:
                legend1 += list(df.columns)
        elif isinstance(df, pd.Series):
            legend1 += [df.name]
        else:
            legend1 += [""]

        min_x_tmp, max_x_tmp = min(df.index), max(df.index)
        if min_x is None or min_x > min_x_tmp:
            min_x = min_x_tmp
        if max_x is None or max_x > max_x_tmp:
            max_x = max_x_tmp

        l1 += ax.plot(df, linestyle=linestyle)
        if y_scales_log[0]:
            ax.set_yscale("log")

    # 分割着色
    if in_sample_date_line is not None:
        if min_x is None:
            logger.warning('min_x is None, cannot draw date_line')
        else:
            x1 = pd.to_datetime(in_sample_date_line)
            p = plt.axvspan(min_x, x1, facecolor=ALTER_BG_COLOR, alpha=0.3)

    if df2 is not None:
        ax2 = ax.twinx()
        ax2.set_prop_cycle(color=get_cmap('Set2').colors)
        if isinstance(df2, pd.DataFrame):
            if df2.shape[1] == 1:
                legend2 = [df2.columns[0]]
            else:
                legend2 = list(df2.columns)
        elif isinstance(df2, pd.Series):
            legend2 = [df2.name]
        else:
            legend2 = [""]
        l2 = ax2.plot(df2, linestyle='--')
        if y_scales_log[1]:
            ax2.set_yscale("log")
        # 设置 legend
        lns = l1 + l2
        legend = legend1 + legend2
    else:
        lns = l1
        legend = legend1

    # legend 参数设置
    # https://blog.csdn.net/helunqu2017/article/details/78641290
    # loc 0 best 3 lower left
    plt.legend(lns, legend, loc=3, frameon=False, ncol=2, fontsize='x-small')  # , framealpha=0.3
    plt.grid(True, color="gray", linewidth="0.5", alpha=0.3)  # , linestyle="-."
    frame = plt.gca()
    frame.axes.get_xaxis().set_visible(True)
    plt.xlim([min_x, max_x])  # x轴边界
    # 设置 title
    # plt.suptitle(name)
    plt.title(name)
    # 展示
    return plot_or_show(enable_save_plot=enable_save_plot, enable_show_plot=enable_show_plot, do_clr=do_clr,
                        file_name=f'{name}.png', folder_path=folder_path)


def _test_plot_twin():
    """测试 plot_twin"""
    date_arr = pd.date_range(pd.to_datetime('2018-01-01'),
                             pd.to_datetime('2018-01-01') + pd.Timedelta(days=99))
    date_index = pd.DatetimeIndex(date_arr)
    close_df = pd.DataFrame({'close': np.sin(np.linspace(0, 10, 100))}, index=date_index)
    accuracy_df = pd.DataFrame({
        'acc1': np.cos(np.linspace(0, 10, 100)) + 9,
        'acc2': np.cos(np.linspace(1, 11, 100)) + 9
    }, index=date_index)
    accuracy2_df = pd.DataFrame({
        'acc_a': (np.cos(np.linspace(0, 10, 100)) + 2) * 10 - 5,
        'acc_b': np.cos(np.linspace(1, 11, 100)) + 9
    }, index=date_index)
    plot_twin([accuracy_df, accuracy2_df], close_df['close'], name='plot twin test log',
              y_scales_log=[True, False], in_sample_date_line='2018-02-01')


@lru_cache()
def get_font_properties():
    is_win = is_windows_os()
    if is_win:
        # 调用系统字体  C:\WINDOWS\Fonts
        font = FontProperties(fname=r"C:\\WINDOWS\\Fonts\\FZSTK.TTF", size=14)
    else:
        # 可以通过 from ibats_utils.mess import get_chinese_font_iter 获取系统有效的中文字体
        # 然后在 /usr/share/fonts/truetype 中查找，
        # 例如：Droid Sans Fallback 字体，对应路径 droid/DroidSansFallbackFull.ttf
        font = FontProperties(fname='/usr/share/fonts/truetype/droid/DroidSansFallbackFull.ttf', size=14)

    return font


def _test_plot_accuracy(single_figure=True):
    """测试 plot_accuracy"""
    date_arr = pd.date_range(pd.to_datetime('2018-01-01'),
                             pd.to_datetime('2018-01-01') + pd.Timedelta(days=99))
    date_index = pd.DatetimeIndex(date_arr)
    close_df = pd.DataFrame({'close': np.sin(np.linspace(0, 10, 100))}, index=date_index)
    accuracy_df = pd.DataFrame({'acc': np.cos(np.linspace(0, 10, 100))}, index=date_index)
    real_pred_df = pd.DataFrame({
        'pred': close_df.apply(lambda x: 1 if x.iloc[0] > 0.3 else 2 if x.iloc[0] < -0.3 else 0, axis=1),
        'real': accuracy_df.apply(lambda x: 1 if x.iloc[0] > 0.3 else 2 if x.iloc[0] < -0.3 else 0, axis=1),
    })
    split_point_list = np.random.randint(len(date_arr), size=10)
    split_point_list.sort()
    split_point_list = date_arr[split_point_list]
    base_line_list = [0.3, 0.6]
    if single_figure:
        enable_kwargs = dict(enable_save_plot=False, enable_show_plot=False, do_clr=False, stg_run_id=6)
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(8, 12))
        ax = fig.add_subplot(211)
        plot_accuracy(accuracy_df, close_df, ax=ax, name='测试使用', split_point_list=split_point_list,
                      base_line_list=base_line_list, **enable_kwargs)
        ax = fig.add_subplot(212)
        plot_accuracy(accuracy_df, close_df, ax=ax, name='测试使用',
                      base_line_list=base_line_list, real_pred_df=real_pred_df, **enable_kwargs)
        plot_or_show(enable_show_plot=True, enable_save_plot=True, file_name='测试使用all in one', stg_run_id=6)
    else:
        plot_accuracy(accuracy_df, close_df, name='测试使用2', split_point_list=split_point_list,
                      base_line_list=base_line_list, stg_run_id=6)


def get_data_range_iter(s: pd.Series, extent_left=False):
    """
    从序列数据中迭代输出每一段相同数据的index范围
    :param s:
    :param extent_left: 左边界与上一个迭代的右边界使用同一个值
    :return:
    """
    is_new_range, idx_from, idx_to, data = True, s.index[0], None, None
    for (idx_to, data), (_, d2) in zip(s.items(), s.shift(-1).items()):
        if is_new_range and not extent_left:
            idx_from = idx_to
            is_new_range = False
        if data != d2 and not (np.isnan(data) and np.isnan(d2)):
            yield idx_from, idx_to, data
            if extent_left:
                idx_from = idx_to
            is_new_range = True
    else:
        if not is_new_range:
            yield idx_from, idx_to, data


if __name__ == "__main__":
    # _test_plot_rr_df()
    # _test_wave_hist()
    # _test_hist_n_rr()
    # _test_label_distribution()
    # _test_n_days_rr_distribution()
    _test_plot_accuracy()
    # _test_show_dl_accuracy()
    # _test_plot_twin()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2019/6/28 17:08
@File    : account.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import numpy as np
import pandas as pd

VERSION_V1 = 'v1'
VERSION_V2 = 'v2'


class Account(object):
    def __init__(self, md_df, data_factors, expand_dims=True, state_with_flag=False, version=VERSION_V1, **kwargs):
        if version == VERSION_V1:
            from ibats_common.backend.rl.emulator.market import QuotesMarket
        elif version == VERSION_V2:
            from ibats_common.backend.rl.emulator.market2 import QuotesMarket
        else:
            raise ValueError(f'param version can only be one of {(VERSION_V1, VERSION_V2)}')
        self.A = QuotesMarket(md_df, data_factors, state_with_flag=state_with_flag, **kwargs)
        self.buffer_reward = []
        self.buffer_value = []
        self.buffer_action = []
        self.buffer_cash = []
        self.buffer_fee = []
        self.buffer_value_fee0 = []
        self.buffer_fee_tot = []
        self.buffer_action_count = []
        self.expand_dims = expand_dims
        self.actions = self.A.get_action_operations()
        self.action_size = len(self.actions)
        self.state_with_flag = state_with_flag

    def reset(self):
        self.buffer_reward = []
        self.buffer_value = []
        self.buffer_action = []
        self.buffer_cash = []
        self.buffer_fee = []
        self.buffer_value_fee0 = []
        self.buffer_fee_tot = []
        self.buffer_action_count = []
        if self.expand_dims:
            # return np.expand_dims(self.A.reset(), 0)
            state = self.A.reset()
            if self.state_with_flag:
                return np.expand_dims(state[0], 0), state[1]
            else:
                return np.expand_dims(state, 0)
        else:
            return self.A.reset()

    def latest_state(self):
        if self.expand_dims:
            # return np.expand_dims(self.A.latest_state(), 0)
            state = self.A.latest_state()
            if self.state_with_flag:
                return np.expand_dims(state[0], 0), state[1]
            else:
                return np.expand_dims(state, 0)
        else:
            return self.A.latest_state()

    def step(self, action):
        next_state, reward, done = self.A.step(action)

        self.buffer_action.append(action)
        self.buffer_reward.append(reward)
        self.buffer_value.append(self.A.total_value)
        self.buffer_cash.append(self.A.cash)
        self.buffer_fee_tot.append(self.A.fee_tot)
        self.buffer_value_fee0.append(self.A.total_value_fee0)
        self.buffer_action_count.append(self.A.action_count)
        if self.expand_dims:
            if self.state_with_flag:
                return (np.expand_dims(next_state[0], 0), next_state[1]), reward, done
            else:
                return np.expand_dims(next_state, 0), reward, done
        else:
            return next_state, reward, done

    def plot_data(self) -> pd.DataFrame:
        reward_df = pd.DataFrame(
            {"value": self.buffer_value,
             "reward": self.buffer_reward,
             "cash": self.buffer_cash,
             "action": self.buffer_action,
             "open": self.A.data_open.iloc[:len(self.buffer_action)],
             "close": self.A.data_close.iloc[:len(self.buffer_action)],
             "fee_tot": self.buffer_fee_tot,
             "value_fee0": self.buffer_value_fee0,
             "action_count": self.buffer_action_count,
             }, index=self.A.data_close.index[:len(self.buffer_action)])
        cum_rr_s = reward_df['value'] / self.A.init_cash - 1
        reward_df['nav'] = cum_linear_rr_2_cum_exp_rr(cum_rr_s)
        cum_rr_s = reward_df['value_fee0'] / self.A.init_cash - 1
        reward_df['nav_fee0'] = cum_linear_rr_2_cum_exp_rr(cum_rr_s)
        return reward_df


def cum_linear_rr_2_cum_exp_rr(cum_rr_s: pd.Series):
    """将 累计线性增长率 曲线转化为 累计指数增长率 曲线"""
    rr = cum_rr_s.copy()
    rr[1:] -= rr.shift(1)[1:]
    rr = rr + 1
    cum_exp_rr = rr.cumprod()
    return cum_exp_rr


def _test_account():
    """测试 env.reset() 返回状态 是否符合预期"""
    n_step = 60
    ohlcav_col_name_list = ["open", "high", "low", "close", "amount", "volume"]
    from ibats_common.example.data import load_data
    md_df = load_data('RB.csv').set_index('trade_date')[ohlcav_col_name_list]
    md_df.index = pd.DatetimeIndex(md_df.index)
    from ibats_common.backend.factor import get_factor, transfer_2_batch
    factors_df = get_factor(md_df, dropna=True)
    df_index, df_columns, data_arr_batch = transfer_2_batch(factors_df, n_step=n_step)
    md_df = md_df.loc[df_index, :]
    shape = [data_arr_batch.shape[0], 5, int(n_step / 5), data_arr_batch.shape[2]]
    data_factors = np.transpose(data_arr_batch.reshape(shape), [0, 2, 3, 1])
    print(data_arr_batch.shape, '->', shape, '->', data_factors.shape)
    # 建立 Account
    env = Account(md_df, data_factors)
    next_observation = env.reset()
    print('next_observation.shape:', next_observation.shape)
    assert next_observation.shape == (1, 12, 79, 5)
    next_state, reward, done = env.step(1)
    assert next_observation.shape == (1, 12, 79, 5)
    assert not done


def _test_account2():
    """测试 plot_data 返回数据是否符合预期"""
    n_step = 60
    ohlcav_col_name_list = ["open", "high", "low", "close", "amount", "volume"]
    from ibats_common.example.data import load_data
    md_df = load_data('RB.csv').set_index('trade_date')[ohlcav_col_name_list]
    md_df.index = pd.DatetimeIndex(md_df.index)
    from ibats_common.backend.factor import get_factor, transfer_2_batch
    factors_df = get_factor(md_df, dropna=True)
    df_index, df_columns, data_arr_batch = transfer_2_batch(factors_df, n_step=n_step)
    md_df = md_df.loc[df_index, :]
    shape = [data_arr_batch.shape[0], 5, int(n_step / 5), data_arr_batch.shape[2]]
    data_factors = np.transpose(data_arr_batch.reshape(shape), [0, 2, 3, 1])
    print(data_arr_batch.shape, '->', shape, '->', data_factors.shape)
    # 建立 Account
    env = Account(md_df, data_factors)
    next_observation = env.reset()
    # 做空
    env.step(2)
    for n in range(int(md_df.shape[0] / 2)):
        env.step(3)
    # 做多
    next_observation, reward, done = env.step(1)
    while not done:
        next_observation, reward, done = env.step(3)

    # 展示结果
    reward_df = env.plot_data()
    value_s = reward_df.iloc[:, 0]
    from ibats_utils.mess import datetime_2_str
    from datetime import datetime
    dt_str = datetime_2_str(datetime.now(), '%Y-%m-%d %H_%M_%S')
    title = f'test_account_{dt_str}'
    from ibats_common.analysis.plot import plot_twin
    plot_twin(value_s, md_df["close"], name=title)


if __name__ == "__main__":
    _test_account()
    _test_account2()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/11/7 10:42
@File    : __init__.py.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import logging
import os
import re
from collections import defaultdict
from datetime import datetime, timedelta

import ffn
import numpy as np
import pandas as pd
from ibats_utils.mess import copy_module_file_to, date_2_str, str_2_date, get_last, copy_file_to
from sklearn.model_selection import train_test_split
from ibats_common.backend.mess import get_report_folder_path
from ibats_common.analysis.plot import show_dl_accuracy
from ibats_common.analysis.summary import summary_release_2_docx
from ibats_common.backend.factor import get_factor
from ibats_common.backend.label import calc_label3
from ibats_common.common import ContextKey, Direction
from ibats_common.example.data import get_trade_date_series, get_delivery_date_series
from ibats_common.strategy import StgBase

logger = logging.getLogger(__name__)
FFN_VERSION = ffn.__version__


class AIStgBase(StgBase):

    def __init__(self, instrument_type, unit=1):
        super().__init__()
        self.unit = unit
        # 模型相关参数
        self.input_size = 38
        self.batch_size = 512
        self.n_step = 60
        self.output_size = 3
        self.n_hidden_units = 72
        self.lr = 0.006
        # 模型训练，及数据集相关参数
        self._model = None
        self._session = None
        self.train_validation_rate = 0.8
        self.xs_train, self.xs_validation, self.ys_train, self.ys_validation = None, None, None, None
        self.label_func_max_rr = 0.0051
        self.label_func_min_rr = -0.0054
        self.max_future = 3
        self.predict_test_random_state = None
        self.n_epoch = 40
        self.max_loop_4_futher_train = 20
        self.retrain_period = 30  # 60 每隔60天重新训练一次，0 则为不进行重新训练
        self.validation_accuracy_base_line = 0.55  # 0.6    # 如果为 None，则不进行 validation 成功率检查
        self.over_fitting_train_acc = 0.95  # 过拟合训练集成功率，如果为None则不进行判断
        # 其他辅助信息
        self.trade_date_series = get_trade_date_series()
        self.delivery_date_series = get_delivery_date_series(instrument_type)
        self.tensorboard_verbose = 3
        # 模型保存路径相关参数
        # 是否使用现有模型进行操作，如果是记得修改以下下方的路径
        # enable_load_model_if_exist 将会在调用 self.load_model_if_exist 时进行检查
        # 如果该字段为 False，调用 load_model_if_exist 时依然可以传入参数的方式加载已有模型
        # 该字段与 self.load_model_if_exist 函数的 enable_load_model_if_exist参数是 “or” 的关系
        self.enable_load_model_if_exist = False
        if self.enable_load_model_if_exist:
            self.base_folder_path = folder_path = os.path.join(
                get_report_folder_path(self.stg_run_id), 'tf_saves_2019-06-27_16_24_34')
        else:
            datetime_str = datetime.now().strftime('%Y-%m-%d_%H_%M_%S')
            self.base_folder_path = folder_path = os.path.join(
                get_report_folder_path(self.stg_run_id), f'tf_saves_{datetime_str}')
        # 备份文件
        file_path = copy_module_file_to(self.__class__, folder_path)
        self.logger.debug('文件已经备份到：%s', file_path)
        # 设置模型备份目录
        self.model_folder_path = model_folder_path = os.path.join(folder_path, 'model_tfls')
        if not os.path.exists(model_folder_path):
            os.makedirs(model_folder_path, exist_ok=True)
        # 设置日至备份目录
        self.tensorboard_dir = tensorboard_dir = os.path.join(folder_path, f'tensorboard_logs')
        if not os.path.exists(tensorboard_dir):
            os.makedirs(tensorboard_dir)
        self.trade_date_last_train = None
        self.trade_date_acc_list = defaultdict(lambda: [0.0, 0.0])
        self.do_nothing_on_min_bar = False  # 仅供调试使用
        # 用于记录 open,high,low,close,amount 的 column 位置
        self.ohlcav_col_name_list = ["open", "high", "low", "close", "amount", "volume"]

    @property
    def session(self):
        return self.get_session()

    def get_session(self, renew=False, close_last_one_if_renew=True):
        import tensorflow as tf
        if renew or self._session is None:
            if self.model is None:
                raise ValueError('model 需要先于 session 被创建')
            if close_last_one_if_renew and self._session is not None:
                self._session.close()
            self._session = tf.Session()
            self._session.run(tf.global_variables_initializer())
        return self._session

    @property
    def model(self):
        """
        获取模型对象
        2019-07-05 去除  -> tflearn.models.DNN hint 该语句将会导致动态加载 class 时抛出错误
        :return:
        """
        return self.get_model()

    def get_model(self, rebuild_model=False):
        """
        获取模型对象，默认不重新构造
        2019-07-05 去除  -> tflearn.models.DNN hint 该语句将会导致动态加载 class 时抛出错误
        :param rebuild_model:
        :return:
        """
        import tensorflow as tf
        if self._model is None or rebuild_model:
            self.logger.info('重新构建模型')
            tf.reset_default_graph()
            self._model = self._build_model()
        return self._model

    def get_factor_array(self, md_df: pd.DataFrame, tail_n=None):
        if tail_n is not None:
            md_df = md_df.tail(tail_n)
        df = md_df[~md_df['close'].isnull()]

        factors = df.fillna(0).to_numpy()
        if self.input_size is None or self.input_size != factors.shape[1]:
            self.input_size = factors.shape[1]
            self.n_hidden_units = self.input_size * 2
            self.logger.info("set input_size: %d", self.input_size)
            self.logger.info("set n_hidden_units: %d", self.n_hidden_units)

        return factors, df.index

    def get_x_y(self, factor_df):
        factors, trade_date_index = self.get_factor_array(factor_df)
        price_arr = factors[:, 0]
        self.input_size = factors.shape[1]
        # ys_all = self.calc_y_against_future_data(price_arr, -0.01, 0.01)
        # ys_all = self.calc_y_against_future_data(price_arr, -self.classify_wave_rate, self.classify_wave_rate)
        ys_all = calc_label3(price_arr, self.label_func_min_rr, self.label_func_max_rr,
                             max_future=self.max_future, one_hot=True)
        range_from = self.n_step
        range_to = factors.shape[0]

        xs = np.zeros((range_to - range_from, self.n_step, self.input_size))
        for num, index in enumerate(range(range_from, range_to)):
            xs[num, :, :] = factors[(index - self.n_step):index, :]

        ys = ys_all[range_from:range_to, :]

        return xs, ys, trade_date_index[range_from:range_to]

    def get_batch_xs(self, factors: np.ndarray, index=None):
        """
        取 batch_xs
        :param factors:
        :param index: 样本起始坐标，如果为None，则默认取尾部一组样本
        :return:
        """
        if index is None:
            index = factors.shape[0] - 1

        batch_xs = np.zeros((1, self.n_step, self.input_size))
        batch_xs[0, :, :] = factors[(index - self.n_step + 1):(index + 1), :]

        return batch_xs

    def _build_model(self):
        # Network building
        # net = tflearn.input_data([None, self.n_step, self.input_size])
        # ...
        # Training
        # _model = tflearn.DNN(net, tensorboard_verbose=self.tensorboard_verbose,
        #                      checkpoint_path=self.checkpoint_path,
        #                      tensorboard_dir=self.tensorboard_dir)
        # return _model
        raise NotImplementedError()

    def train(self, factor_df_dic: dict, predict_test_random_state):
        import tflearn
        factor_df = factor_df_dic[1]

        trade_date_from_str, trade_date_to_str = date_2_str(factor_df.index[0]), date_2_str(factor_df.index[-1])
        # xs_train, xs_validation, ys_train, ys_validation = self.separate_train_validation(xs, ys)
        if self.predict_test_random_state is None:
            random_state = predict_test_random_state
        else:
            random_state = self.predict_test_random_state

        # 利用生成数据做训练数据集，只用原始数据中的 validation 部分做验证集
        arr_list, xs_validation, ys_validation = [], None, None
        for adj_factor, factor_df in factor_df_dic.items():
            xs, ys, _ = self.get_x_y(factor_df)
            xs_train_tmp, xs_validation_tmp, ys_train_tmp, ys_validation_tmp = train_test_split(
                xs, ys, test_size=0.2, random_state=random_state)
            arr_list.append([xs_train_tmp, ys_train_tmp])
            # xs_train, xs_validation = xs_train_tmp, xs_validation_tmp
            # ys_train, ys_validation = ys_train_tmp, ys_validation_tmp
            if adj_factor == 1:
                xs_validation, ys_validation = xs_validation_tmp, ys_validation_tmp

        self.xs_train = xs_train = np.vstack([_[0] for _ in arr_list])
        self.ys_train = ys_train = np.vstack([_[1] for _ in arr_list])

        sess = self.get_session(renew=True)
        train_acc, val_acc = 0, 0
        with sess.as_default():
            # with tf.Graph().as_default():
            # self.logger.debug('sess.graph:%s tf.get_default_graph():%s', sess.graph, tf.get_default_graph())
            self.logger.debug('[%d], xs_train %s, ys_train %s, xs_validation %s, ys_validation %s, [%s, %s]',
                              random_state, xs_train.shape, ys_train.shape, xs_validation.shape, ys_validation.shape,
                              trade_date_from_str, trade_date_to_str)
            max_loop = self.max_loop_4_futher_train
            for num in range(max_loop):
                if num == 0:
                    n_epoch = self.n_epoch
                else:
                    n_epoch = self.n_epoch // max_loop

                self.logger.info('[%d]第 %d/%d 轮训练，开始 [%s, %s] n_epoch=%d', random_state, num + 1, max_loop,
                                 trade_date_from_str, trade_date_to_str, n_epoch)
                run_id = f'{trade_date_to_str}_{xs_train.shape[0]}[{predict_test_random_state}]' \
                         f'_{datetime.now().strftime("%Y%m%d_%H%M%S")}'
                tflearn.is_training(True)
                self.model.fit(
                    xs_train, ys_train, validation_set=(xs_validation, ys_validation),
                    show_metric=True, batch_size=self.batch_size, n_epoch=n_epoch,
                    run_id=run_id)
                tflearn.is_training(False)

                result = self.model.evaluate(xs_train, ys_train, batch_size=self.batch_size)
                # self.logger.info("train accuracy: %.2f%%" % (result[0] * 100))
                train_acc = result[0]
                result = self.model.evaluate(xs_validation, ys_validation, batch_size=self.batch_size)
                val_acc = result[0]
                self.logger.info("[%d]第 %d/%d 轮训练，[%s - %s]，训练集准确率(train_acc)：%.2f%%， 样本外准确率(val_acc): %.2f%%",
                                 random_state, num + 1, max_loop, trade_date_from_str, trade_date_to_str,
                                 train_acc * 100, val_acc * 100)
                if self.over_fitting_train_acc is not None and train_acc > self.over_fitting_train_acc:
                    self.logger.warning('[%d]第 %d/%d 轮训练，训练集精度 %.2f%% > %.2f%% 可能存在过拟合 [%s, %s]',
                                        random_state, num + 1, max_loop, train_acc * 100,
                                        self.over_fitting_train_acc * 100,
                                        trade_date_from_str, trade_date_to_str)
                    break
                if self.validation_accuracy_base_line is not None:
                    if result[0] > self.validation_accuracy_base_line:
                        break
                    elif num < max_loop - 1:
                        self.logger.warning('[%d]第 %d/%d 轮训练，[%s - %s]，样本外训练准确率 %.2f%% < %.0f%%，继续训练',
                                            random_state, num + 1, max_loop, trade_date_from_str, trade_date_to_str,
                                            val_acc * 100, self.validation_accuracy_base_line * 100)
                else:
                    break

        self.trade_date_last_train = str_2_date(trade_date_to_str)
        return train_acc, val_acc

    def valid_model_acc(self, factor_df: pd.DataFrame):
        xs, ys, _ = self.get_x_y(factor_df)
        trade_date_from_str, trade_date_to_str = date_2_str(factor_df.index[0]), date_2_str(factor_df.index[-1])
        random_state = self.predict_test_random_state
        xs_train, xs_validation, ys_train, ys_validation = train_test_split(
            xs, ys, test_size=0.2, random_state=random_state)
        self.logger.debug('random_state=%d, xs_train %s, ys_train %s, xs_validation %s, ys_validation %s, [%s, %s]',
                          random_state, xs_train.shape, ys_train.shape, xs_validation.shape, ys_validation.shape,
                          trade_date_from_str, trade_date_to_str)
        sess = self.get_session(renew=True)
        with sess.as_default():
            # with tf.Graph().as_default():
            # self.logger.debug('sess.graph:%s tf.get_default_graph():%s', sess.graph, tf.get_default_graph())
            result = self.model.evaluate(xs_validation, ys_validation, batch_size=self.batch_size)
            val_acc = result[0]
            result = self.model.evaluate(xs_train, ys_train, batch_size=self.batch_size)
            train_acc = result[0]
            self.logger.info("[%s - %s] 训练集准确率: %.2f%%", trade_date_from_str, trade_date_to_str, train_acc * 100)
            self.logger.info("[%s - %s] 验证集准确率: %.2f%%", trade_date_from_str, trade_date_to_str, val_acc * 100)
        return train_acc, val_acc

    def save_model(self, trade_date):
        """
        将模型导出到文件
        :param trade_date:
        :return:
        """
        folder_path = os.path.join(self.model_folder_path, date_2_str(trade_date))
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        file_path = os.path.join(
            folder_path,
            f"model_{int(self.label_func_min_rr * 10000)}_{int(self.label_func_max_rr * 10000)}.tfl")
        self.model.save(file_path)
        self.logger.info("模型训练截止日期： %s 保存到: %s", self.trade_date_last_train, file_path)
        return file_path

    def load_model_if_exist(self, trade_date, enable_load_model_if_exist=False):
        """
        将模型导出到文件
        目录结构：
        tf_saves_2019-06-05_16_21_39
          *   model_tfls
          *       *   2012-12-31
          *       *       *   checkpoint
          *       *       *   model_-54_51.tfl.data-00000-of-00001
          *       *       *   model_-54_51.tfl.index
          *       *       *   model_-54_51.tfl.meta
          *       *   2013-02-28
          *       *       *   checkpoint
          *       *       *   model_-54_51.tfl.data-00000-of-00001
          *       *       *   model_-54_51.tfl.index
          *       *       *   model_-54_51.tfl.meta
          *   tensorboard_logs
          *       *   2012-12-31_496[1]_20190605_184316
          *       *       *   events.out.tfevents.1559731396.mg-ubuntu64
          *       *   2013-02-28_496[1]_20190605_184716
          *       *       *   events.out.tfevents.1559731396.mg-ubuntu64
        :param enable_load_model_if_exist:
        :param trade_date:
        :return:
        """
        if self.enable_load_model_if_exist or enable_load_model_if_exist:
            # 获取小于等于当期交易日的最大的一个交易日对应的文件名
            min_available_date = str_2_date(trade_date) - timedelta(days=self.retrain_period)
            self.logger.debug('尝试加载现有模型，[%s - %s] %d 天', min_available_date, trade_date, self.retrain_period)
            date_file_path_pair_list = [_ for _ in self.get_date_file_path_pair_list() if _[0] >= min_available_date]
            if len(date_file_path_pair_list) > 0:
                # 按日期排序
                date_file_path_pair_list.sort(key=lambda x: x[0])
                # 获取小于等于当期交易日的最大的一个交易日对应的文件名
                # file_path = get_last(date_file_path_pair_list, lambda x: x[0] <= trade_date, lambda x: x[1])
                trade_date = str_2_date(trade_date)
                ret = get_last(date_file_path_pair_list, lambda x: x[0] <= trade_date)
                if ret is not None:
                    key, folder_path, predict_test_random_state = ret
                    if folder_path is not None:
                        model = self.get_model(rebuild_model=True)  # 这句话是必须的，需要实现建立模型才可以加载
                        model.load(folder_path)
                        self.trade_date_last_train = key
                        self.predict_test_random_state = predict_test_random_state
                        self.logger.info("加载模型成功。trade_date_last_train: %s load from path: %s", key, folder_path)
                        return True

        return False

    def predict_test(self, md_df):
        self.logger.info('开始预测')
        # sess = self.session
        if self.xs_train is None:
            xs, ys, _ = self.get_x_y(md_df)
            xs_train, xs_validation, ys_train, ys_validation = train_test_split(xs, ys, test_size=0.2, random_state=1)
            self.xs_train, self.xs_validation, self.ys_train, self.ys_validation = xs_train, xs_validation, ys_train, ys_validation

        with self.session.as_default() as sess:
            real_ys = np.argmax(self.ys_validation, axis=1)

            # self.logger.info("批量预测2")
            # pred_ys = np.argmax(self.model.predict_label(self.xs_validation), axis=1) 与 evaluate 结果刚好相反
            # 因此不使用 predict_label 函数
            # pred_ys = np.argmax(self.model.predict(self.xs_validation), axis=1)
            # self.logger.info("模型训练基准日期：%s，validation accuracy: %.2f%%",
            #             self.trade_date_last_train, sum(pred_ys == real_ys) / len(pred_ys) * 100)
            # self.logger.info("pred/real: \n%s\n%s", pred_ys, real_ys)

            self.logger.info("独立样本预测(predict_latest)")
            pred_ys = []
            for idx, y in enumerate(self.ys_validation):
                x = self.xs_validation[idx:idx + 1, :, :]
                pred_y = self.model.predict(x)
                pred_ys.extend(np.argmax(pred_y, axis=1))

            pred_ys = np.array(pred_ys)
            self.logger.info("模型训练基准日期：%s，validation accuracy: %.2f%%",
                             self.trade_date_last_train, sum(pred_ys == real_ys) / len(pred_ys) * 100)
            # self.logger.info("pred: \n%s\n%s", pred_ys, real_ys)

    def predict_latest(self, factor_df):
        """
        计算最新一个 X，返回分类结果
        二分类，返回 0 未知 / 1 下跌 / 2 上涨
        三分类，返回 0 震荡 / 1 下跌 / 2 上涨
        :param factor_df:
        :return:
        """
        factors, _ = self.get_factor_array(factor_df, tail_n=self.n_step)
        x = self.get_batch_xs(factors)
        pred_y = np.argmax(self.model.predict(x), axis=1)[-1]
        # is_buy, is_sell = pred_mark == 1, pred_mark == 0
        # return is_buy, is_sell
        return pred_y

    def on_prepare_min1(self, md_df, context):
        if md_df is None:
            return
        indexed_df = md_df.set_index('trade_date').drop('instrument_type', axis=1)
        indexed_df.index = pd.DatetimeIndex(indexed_df.index)
        self.load_train_test(indexed_df, enable_load_model=self.enable_load_model_if_exist)

    def load_train_test(self, indexed_df, enable_load_model, rebuild_model=False, enable_train_if_load_not_suss=True,
                        enable_train_even_load_succ=False, enable_test=False):
        if rebuild_model:
            self.get_model(rebuild_model=True)

        trade_date = str_2_date(indexed_df.index[-1])
        # 加载模型
        if enable_load_model:
            is_load = self.load_model_if_exist(trade_date)
        else:
            is_load = False

        if enable_train_even_load_succ or (enable_train_if_load_not_suss and not is_load):
            factor_df_dic = get_factor(indexed_df, ohlcav_col_name_list=self.ohlcav_col_name_list,
                                       trade_date_series=self.trade_date_series,
                                       delivery_date_series=self.delivery_date_series, do_multiple_factors=True)
            factor_df = factor_df_dic[1]
            num = 0
            while True:
                num += 1
                if num > 1:
                    self.get_model(rebuild_model=True)
                # 训练模型
                train_acc, val_acc = self.train(factor_df_dic, predict_test_random_state=num)
                if self.over_fitting_train_acc is not None and train_acc > self.over_fitting_train_acc:
                    self.logger.warning('第 %d 次训练，训练集精度 train_acc=%.2f%% 过高，可能存在过拟合，重新采样训练',
                                        num, train_acc * 100)
                    continue
                if self.validation_accuracy_base_line is not None:
                    if val_acc < self.validation_accuracy_base_line:
                        self.logger.warning('第 %d 次训练，训练结果不及预期，重新采样训练', num)
                        continue
                    # elif train_acc - val_acc > 0.15 and val_acc < 0.75:
                    #     self.logger.warning('第 %d 次训练，train_acc=%.2f%%, val_acc=%.2f%% 相差大于 15%% 且验证集正确率小于75%%，重新采样训练',
                    #                    num, train_acc * 100, val_acc * 100)
                    #     continue
                    else:
                        break
                else:
                    break

            self.save_model(trade_date)
            self.trade_date_last_train = trade_date
        else:
            factor_df = get_factor(indexed_df, ohlcav_col_name_list=self.ohlcav_col_name_list,
                                   trade_date_series=self.trade_date_series,
                                   delivery_date_series=self.delivery_date_series)
            train_acc, val_acc = self.valid_model_acc(factor_df)

        self.trade_date_acc_list[trade_date] = [train_acc, val_acc]

        # enable_test 默认为 False
        # self.valid_model_acc(factor_df) 以及完全取代 self.predict_test
        # self.predict_test 仅用于内部测试使用
        if enable_test:
            self.predict_test(factor_df)

        return factor_df

    def on_min1(self, md_df, context):
        if self.do_nothing_on_min_bar:  # 仅供调试使用
            return

        # 数据整理
        indexed_df = md_df.set_index('trade_date').drop('instrument_type', axis=1)
        indexed_df.index = pd.DatetimeIndex(indexed_df.index)
        # 获取最新交易日
        trade_date = str_2_date(indexed_df.index[-1])
        days_after_last_train = (trade_date - self.trade_date_last_train).days
        if self.retrain_period is not None and 0 < self.retrain_period < days_after_last_train:
            # 重新训练
            self.logger.info('当前日期 %s 距离上一次训练 %s 已经过去 %d 天，重新训练',
                             trade_date, self.trade_date_last_train, days_after_last_train)
            factor_df = self.load_train_test(indexed_df, rebuild_model=True,
                                             enable_load_model=self.enable_load_model_if_exist)
        else:
            factor_df = get_factor(indexed_df, ohlcav_col_name_list=self.ohlcav_col_name_list,
                                   trade_date_series=self.trade_date_series,
                                   delivery_date_series=self.delivery_date_series)

        # 预测
        pred_mark = self.predict_latest(factor_df)
        is_holding, is_buy, is_sell = pred_mark == 0, pred_mark == 1, pred_mark == 2
        # self.logger.info('%s is_buy=%s, is_sell=%s', trade_date, str(is_buy), str(is_sell))
        close = md_df['close'].iloc[-1]
        instrument_id = context[ContextKey.instrument_id_list][0]
        if is_buy:  # is_buy
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Short:
                        self.close_short(instrument_id, close, pos_info.position)
                    elif direction == Direction.Long:
                        no_target_position = False
            if no_target_position:
                self.open_long(instrument_id, close, self.unit)
            else:
                self.logger.debug("%s %s     %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

        if is_sell:  # is_sell
            position_date_pos_info_dic = self.get_position(instrument_id)
            no_holding_target_position = True
            if position_date_pos_info_dic is not None:
                for position_date, pos_info in position_date_pos_info_dic.items():
                    direction = pos_info.direction
                    if direction == Direction.Long:
                        self.close_long(instrument_id, close, pos_info.position)
                    elif direction == Direction.Short:
                        no_holding_target_position = False
            if no_holding_target_position:
                self.open_short(instrument_id, close, self.unit)
            else:
                self.logger.debug("%s %s     %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

        if is_holding:
            self.logger.debug("%s %s * * %.2f holding", self.trade_agent.curr_timestamp, instrument_id, close)

    def on_min1_release(self, md_df):
        """
        增加模型对未来数据预测成功率走势图展示
        :param md_df:
        :return:
        """
        if md_df is None or md_df.shape[0] == 0:
            self.logger.warning('md_df is None or shape[0] == 0')
            return
        else:
            self.logger.debug('md_df.shape= %s', md_df.shape)

        # 获取各个模型训练时间点及路径
        date_file_path_pair_list = self.get_date_file_path_pair_list()
        if len(date_file_path_pair_list) > 0:
            # 按日期排序
            date_file_path_pair_list.sort(key=lambda x: x[0])

        # 建立数据集
        indexed_df = md_df.set_index('trade_date').drop('instrument_type', axis=1)
        trade_date_end = indexed_df.index[-1]
        factor_df = get_factor(indexed_df, ohlcav_col_name_list=self.ohlcav_col_name_list,
                               trade_date_series=self.trade_date_series,
                               delivery_date_series=self.delivery_date_series)
        xs, ys_onehot, trade_date_index = self.get_x_y(factor_df)
        ys = np.argmax(ys_onehot, axis=1)
        data_len = len(trade_date_index)
        if data_len == 0:
            self.logger.warning('ys 长度为0，请检查是否存在数据错误')
            return
        trade_date2_list = [_[0] for _ in date_file_path_pair_list][1:]
        trade_date2_list.append(None)
        # 预测结果
        self.logger.info("按日期分段验证检验预测结果")
        pred_ys_tot, real_ys_tot, img_meta_dic_list = [], [], []
        # 根据模型 trade_date_last_train 进行分段预测，并将结果记录到 pred_ys
        for num, ((trade_date_last_train, file_path, predict_test_random_state), trade_date_next) in enumerate(zip(
                date_file_path_pair_list, trade_date2_list)):
            # 以模型训练日期为基准，后面的数据作为验证集数据（样本外数据）
            # 获取有效的日期范围 from - to
            range_from_arr = trade_date_index >= pd.to_datetime(trade_date_last_train)
            range_from_len = len(range_from_arr)
            if range_from_len == 0:  # range_from_len 应该与 trade_date_list_count 等长度，所以这个条件应该永远不会满足
                self.logger.error('总共%d条数据，%s 开始后面没有可验证数据', data_len, trade_date_last_train)
                continue
            true_count = sum(range_from_arr)
            self.logger.debug("len(range_from)=%d, True Count=%d", len(range_from_arr), true_count)
            if true_count == 0:
                self.logger.warning('总共%d条数据，%s 开始后面没有可验证数据', data_len, trade_date_last_train)
                continue
            # 自 trade_date_last_train 起的所有有效日期
            trade_date_list_sub = trade_date_index[range_from_arr]

            # 获取 in_range，作为 range_from, range_to 的交集
            if trade_date_next is None:
                in_range_arr = None
                in_range_count = true_count
            else:
                in_range_arr = trade_date_list_sub < pd.to_datetime(trade_date_next)
                in_range_count = sum(in_range_arr)
                if in_range_count == 0:
                    self.logger.warning('总共%d条数据，[%s - %s) 之间没有可用数据',
                                        data_len, trade_date_last_train, trade_date_next)
                    continue
                else:
                    self.logger.debug('总共%d条数据，[%s - %s) 之间有 %d 条数据将被验证 model path:%s',
                                      data_len, trade_date_last_train, trade_date_next, in_range_count, file_path)

            # 获取当前时段对应的 xs
            # 进行验证时，对 range_from 开始的全部数据进行预测，按照 range_to 为分界线分区着色显示
            xs_sub, real_ys = xs[range_from_arr, :, :], ys[range_from_arr]
            close_df = indexed_df.loc[trade_date_list_sub, 'close']

            # 加载模型
            is_load = self.load_model_if_exist(trade_date_last_train, enable_load_model_if_exist=True)
            if not is_load:
                self.logger.error('%s 模型加载失败：%s', trade_date_last_train, file_path)
                continue
            # 预测
            pred_ys_one_hot = self.model.predict(xs_sub)
            pred_ys = np.argmax(pred_ys_one_hot, axis=1)
            if in_range_arr is not None and in_range_count > 0:
                pred_ys_tot.extend(pred_ys[in_range_arr])
            else:
                pred_ys_tot.extend(pred_ys)

            # 为每一个时段单独验证成功率，以当前模型为基准，验证后面全部历史数据成功率走势
            if trade_date_next is None:
                split_point_list = None
            else:
                split_point_list = [close_df.index[0], trade_date_next, close_df.index[-1]]
            base_line_list = self.trade_date_acc_list[trade_date_last_train]
            img_file_path = show_dl_accuracy(real_ys, pred_ys, close_df, split_point_list,
                                             base_line_list=base_line_list)
            img_meta_dic_list.append({
                'img_file_path': img_file_path,
                'trade_date_last_train': trade_date_last_train,
                'module_file_path': file_path,
                'predict_test_random_state': predict_test_random_state,
                'split_point_list': split_point_list,
                'in_range_count': in_range_count,
                'trade_date_end': trade_date_end,
            })

        pred_ys_tot = np.array(pred_ys_tot)
        trade_date_last_train_first = pd.to_datetime(date_file_path_pair_list[0][0])
        split_point_list = [_[0] for _ in date_file_path_pair_list]
        split_point_list.append(trade_date_index[-1])
        # 获取 real_ys
        real_ys = ys[trade_date_index >= trade_date_last_train_first]
        close_df = indexed_df.loc[trade_date_index[trade_date_index >= trade_date_last_train_first], 'close']
        img_file_path = show_dl_accuracy(real_ys, pred_ys_tot, close_df, split_point_list)

        img_meta_dic_list.append({
            'img_file_path': img_file_path,
            'trade_date_last_train': trade_date_last_train_first,
            'module_file_path': date_file_path_pair_list[0][1],
            'predict_test_random_state': date_file_path_pair_list[0][2],
            'split_point_list': split_point_list,
            'in_range_count': close_df.shape[0],
            'trade_date_end': trade_date_end,
        })
        is_output_docx = True
        if is_output_docx:
            title = f"[{self.stg_run_id}] predict accuracy trend report"
            file_path = summary_release_2_docx(title, img_meta_dic_list)
            copy_file_to(file_path, self.base_folder_path)

    def get_date_file_path_pair_list(self):
        """
        目录结构：
        tf_saves_2019-06-05_16_21_39
          *   model_tfls
          *       *   2012-12-31
          *       *       *   checkpoint
          *       *       *   model_-54_51.tfl.data-00000-of-00001
          *       *       *   model_-54_51.tfl.index
          *       *       *   model_-54_51.tfl.meta
          *       *   2013-02-28
          *       *       *   checkpoint
          *       *       *   model_-54_51.tfl.data-00000-of-00001
          *       *       *   model_-54_51.tfl.index
          *       *       *   model_-54_51.tfl.meta
          *   tensorboard_logs
          *       *   2012-12-31_496[1]_20190605_184316
          *       *       *   events.out.tfevents.1559731396.mg-ubuntu64
          *       *   2013-02-28_496[1]_20190605_184716
          *       *       *   events.out.tfevents.1559731396.mg-ubuntu64
        :return:
        """
        # 获取全部文件名
        pattern = re.compile(r'model_[-]?\d+_\d+.tfl')
        date_file_path_pair_list, model_name_set = [], set()
        for folder_name in os.listdir(self.model_folder_path):
            folder_path = os.path.join(self.model_folder_path, folder_name)
            if os.path.isdir(folder_path):
                try:
                    # 获取 trade_date_last_train
                    key = str_2_date(folder_name)
                    for file_name in os.listdir(folder_path):
                        # 对下列有效文件名，匹配结果："model_-54_51.tfl"
                        # model_-54_51.tfl.data-00000-of-00001
                        # model_-54_51.tfl.index
                        # model_-54_51.tfl.meta
                        m = pattern.search(file_name)
                        if m is None:
                            continue
                        model_name = m.group()
                        if key in model_name_set:
                            continue
                        model_name_set.add(key)
                        # 获取 model folder_path
                        file_path = os.path.join(folder_path, model_name)
                        # 获取 predict_test_random_state
                        for log_folder_path in os.listdir(self.tensorboard_dir):
                            if log_folder_path.find(folder_name) == 0:
                                predict_test_random_state = int(log_folder_path.split('[')[1].split(']')[0])
                                break
                        else:
                            predict_test_random_state = None

                        date_file_path_pair_list.append([key, file_path, predict_test_random_state])
                except:
                    pass

        return date_file_path_pair_list



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/6/26 10:27
@File    : __init__.py.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import pandas as pd

pd.set_option('display.width', 240)
pd.set_option('display.max_columns', 20)
pd.set_option('display.float_format', '{:,.4f}'.format)
